/*****************************************************************************
 *                                                                           *
 SC 29 Software Copyright Licencing Disclaimer:

 This software module was originally developed by
 AT&T, Dolby Laboratories, Fraunhofer IIS-A

 and edited by
  Mitsuyuki Hatanaka (Sony Corporation),
  Toru Chinen        (Sony Corporation)

 in the course of development of the ISO/IEC 13818-7 and ISO/IEC 14496-3
 standards for reference purposes and its performance may not have been
 optimized. This software module is an implementation of one or more tools as
 specified by the ISO/IEC 13818-7 and ISO/IEC 14496-3 standards.
 ISO/IEC gives users free license to this software module or modifications
 thereof for use in products claiming conformance to audiovisual and
 image-coding related ITU Recommendations and/or ISO/IEC International
 Standards. ISO/IEC gives users the same free license to this software module or
 modifications thereof for research purposes and further ISO/IEC standardisation.
 Those intending to use this software module in products are advised that its
 use may infringe existing patents. ISO/IEC have no liability for use of this
 software module or modifications thereof. Copyright is not released for
 products that do not conform to audiovisual and image-coding related ITU
 Recommendations and/or ISO/IEC International Standards.
 The original developer retains full right to modify and use the code for its
 own purpose, assign or donate the code to a third party and to inhibit third
 parties from using the code for products that do not conform to audiovisual and
 image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 This copyright notice must be included in all copies or derivative works.
 Copyright (c) ISO/IEC 1996.
 *                                                                           *
 ****************************************************************************/

#include "config.h"
#include "adifdec.h"
#include "diagnose.h"
#include "huffinit.h"
#include "interface.h"
#include "bitinput.h"
#include "tns.h"

/* --- local data --- */

static int pred_max_bands_tbl[(1 << LEN_SAMP_IDX)] =
{ 33, /* 96000 */
33, /* 88200 */
38, /* 64000 */
40, /* 48000 */
40, /* 44100 */
40, /* 32000 */
41, /* 24000 */
41, /* 22050 */
37, /* 16000 */
37, /* 12000 */
37, /* 11025 */
34, /* 8000  */
34, /* 7350  */
0, 0, 0 };

static int tns_max_bands_tbl[(1 << LEN_SAMP_IDX)][4] =
{
/* entry for each sampling rate
 * 1    Main/LC long window
 * 2    Main/LC short window
 * 3    SSR long window
 * 4    SSR short window
 */
{ 31, 9, 28, 7 }, /* 96000 */
{ 31, 9, 28, 7 }, /* 88200 */
{ 34, 10, 27, 7 }, /* 64000 */
{ 40, 14, 26, 6 }, /* 48000 */
{ 42, 14, 26, 6 }, /* 44100 */
{ 51, 14, 26, 6 }, /* 32000 */
{ 46, 14, 29, 7 }, /* 24000 */
{ 46, 14, 29, 7 }, /* 22050 */
{ 42, 14, 23, 8 }, /* 16000 */
{ 42, 14, 23, 8 }, /* 12000 */
{ 42, 14, 23, 8 }, /* 11025 */
{ 39, 14, 19, 7 }, /* 8000  */
{ 39, 14, 19, 7 }, /* 7350  */
{ 0, 0, 0, 0 },
{ 0, 0, 0, 0 },
{ 0, 0, 0, 0 } };

static int sampling_boundarys[(1 << LEN_SAMP_IDX)] =
{ 92017, 75132, 55426, 46009, 37566, 27713, 23004, 18783, 13856, 11502, 9391,
    0, 0, 0, 0, 0 };

/* --- global data --- */

ProgConfig prog_config;
MC_Info mc_info;
int LN  = LN_MAX; /* needed when ASC missing */
int LN2 = LN2_MAX;
int LN4 = LN4_MAX;
int SN  = SN_MAX; /* needed when ASC missing */
int SN2 = SN2_MAX;
int SN4 = SN4_MAX;


int current_program;
int default_config;
/* Number of channels for decoderBitBuffer-calculations */
int channel_num;

int sampling_rate[(1 << LEN_SAMP_IDX)] =
{ 96000, 88200, 64000, 48000, 44100, 32000, 24000, 22050, 16000, 12000, 11025,
    8000, 7350, 0, 0, 0 };
const char *sfi[] = { 
  "96000 Hz", "88200 Hz", "64000 Hz", "48000 Hz", "44100 Hz", "32000 Hz",
  "24000 Hz", "22050 Hz", "16000 Hz", "12000 Hz", "11025 Hz", "8000 Hz",
  "7350 Hz", NULL, NULL, "escape"

};
const char *profile[] = { "main", "LC", "SSR", "LTP" };
const char *ot[] = { "Null", "AAC main", "AAC LC", "AAC SSR", "AAC LTP", "SBR", "AAC Scalable", "TwinVQ", "CELP", "HVXC", NULL, NULL, "TTSI", "Main synthetic", "Wavetable synthesis", "General MIDI", "Algorithmic Synthesis and Audio FX", "ER AAC LC", NULL, "ER AAC LTP", "ER AAC Scalable", "ER TwinVQ", "ER BSAC", "ER AAC LD", "ER CELP", "ER HVXC", "ER HILN", "ER Parametric", "SSC", "PS", "MPEG Surround", "escape", "Layer-1", "Layer-2", "Layer-3", "DST", "ALS", "SLS", "SLS non-core" };

#ifdef AMD4_S
/* table for CRC8 */
static unsigned char crc8_tbl[ 256 ] = {
	0x00, 0x07, 0x0E, 0x09, 0x1C, 0x1B, 0x12, 0x15,
	0x38, 0x3F, 0x36, 0x31, 0x24, 0x23, 0x2A, 0x2D,
	0x70, 0x77, 0x7E, 0x79, 0x6C, 0x6B, 0x62, 0x65,
	0x48, 0x4F, 0x46, 0x41, 0x54, 0x53, 0x5A, 0x5D,
	0xE0, 0xE7, 0xEE, 0xE9, 0xFC, 0xFB, 0xF2, 0xF5,
	0xD8, 0xDF, 0xD6, 0xD1, 0xC4, 0xC3, 0xCA, 0xCD,
	0x90, 0x97, 0x9E, 0x99, 0x8C, 0x8B, 0x82, 0x85,
	0xA8, 0xAF, 0xA6, 0xA1, 0xB4, 0xB3, 0xBA, 0xBD,
	
	0xC7, 0xC0, 0xC9, 0xCE, 0xDB, 0xDC, 0xD5, 0xD2,
	0xFF, 0xF8, 0xF1, 0xF6, 0xE3, 0xE4, 0xED, 0xEA,
	0xB7, 0xB0, 0xB9, 0xBE, 0xAB, 0xAC, 0xA5, 0xA2,
	0x8F, 0x88, 0x81, 0x86, 0x93, 0x94, 0x9D, 0x9A,
	0x27, 0x20, 0x29, 0x2E, 0x3B, 0x3C, 0x35, 0x32,
	0x1F, 0x18, 0x11, 0x16, 0x03, 0x04, 0x0D, 0x0A,
	0x57, 0x50, 0x59, 0x5E, 0x4B, 0x4C, 0x45, 0x42,
	0x6F, 0x68, 0x61, 0x66, 0x73, 0x74, 0x7D, 0x7A,
	
	0x89, 0x8E, 0x87, 0x80, 0x95, 0x92, 0x9B, 0x9C,
	0xB1, 0xB6, 0xBF, 0xB8, 0xAD, 0xAA, 0xA3, 0xA4,
	0xF9, 0xFE, 0xF7, 0xF0, 0xE5, 0xE2, 0xEB, 0xEC,
	0xC1, 0xC6, 0xCF, 0xC8, 0xDD, 0xDA, 0xD3, 0xD4,
	0x69, 0x6E, 0x67, 0x60, 0x75, 0x72, 0x7B, 0x7C,
	0x51, 0x56, 0x5F, 0x58, 0x4D, 0x4A, 0x43, 0x44,
	0x19, 0x1E, 0x17, 0x10, 0x05, 0x02, 0x0B, 0x0C,
	0x21, 0x26, 0x2F, 0x28, 0x3D, 0x3A, 0x33, 0x34,
	
	0x4E, 0x49, 0x40, 0x47, 0x52, 0x55, 0x5C, 0x5B,
	0x76, 0x71, 0x78, 0x7F, 0x6A, 0x6D, 0x64, 0x63,
	0x3E, 0x39, 0x30, 0x37, 0x22, 0x25, 0x2C, 0x2B,
	0x06, 0x01, 0x08, 0x0F, 0x1A, 0x1D, 0x14, 0x13,
	0xAE, 0xA9, 0xA0, 0xA7, 0xB2, 0xB5, 0xBC, 0xBB,
	0x96, 0x91, 0x98, 0x9F, 0x8A, 0x8D, 0x84, 0x83,
	0xDE, 0xD9, 0xD0, 0xD7, 0xC2, 0xC5, 0xCC, 0xCB,
	0xE6, 0xE1, 0xE8, 0xEF, 0xFA, 0xFD, 0xF4, 0xF3,
};

#endif /* AMD4_S */

#ifdef AMD5_UNIDRC
FILE *fUniDrcData = NULL;
#endif

/* --- functions --- */

/* enter program configuration into MC_Info structure
 *  only configures for channels specified in chandefs.h
 */
static int enter_mc_info(MC_Info *mip, ProgConfig *pcp)
{
  int i, j, id, tag, cw;
  EleList *elp;
  MIXdown *mxp;

  /* reset channel counts */
  mip->nch = 0;
  mip->nfch = 0;
  mip->nfsce = 0;
  mip->nsch = 0;
  mip->nbch = 0;
  mip->nlch = 0;
  mip->ncch = 0;

  /* profile and sampling rate
   *	re-configure if new sampling rate
   */
  mip->profile = pcp->profile;
  if (mip->sampling_rate_idx != pcp->sampling_rate_idx)
  {
    mip->sampling_rate_idx = pcp->sampling_rate_idx;
  }
  mip->sampling_rate = sampling_rate[mip->sampling_rate_idx];

  cw = 0; /* changed later */

  /* front elements, center out */
  elp = &pcp->front;
  /* count number of leading SCE's */
  for (i = 0, j = elp->num_ele; i < j; i++)
  {
    if (elp->ele_is_cpe[i])
      break;
    mip->nfsce++;
  }
  for (i = 0, j = elp->num_ele; i < j; i++)
  {
    id = elp->ele_is_cpe[i] ? ID_CPE : ID_SCE;
    tag = elp->ele_tag[i];
    if (enter_chn(id, tag, 'f', cw, mip) < 0)
      return (-1);
  }

  /* side elements, left to right then front to back */
  elp = &pcp->side;
  for (i = 0, j = elp->num_ele; i < j; i++)
  {
    id = elp->ele_is_cpe[i] ? ID_CPE : ID_SCE;
    tag = elp->ele_tag[i];
    if (enter_chn(id, tag, 's', cw, mip) < 0)
      return (-1);
  }

  /* back elements, outside to center */
  elp = &pcp->back;
  for (i = 0, j = elp->num_ele; i < j; i++)
  {
    id = elp->ele_is_cpe[i] ? ID_CPE : ID_SCE;
    tag = elp->ele_tag[i];
    if (enter_chn(id, tag, 'b', cw, mip) < 0)
      return (-1);
  }

  /* lfe elements */
  elp = &pcp->lfe;
  for (i = 0, j = elp->num_ele; i < j; i++)
  {
    id = elp->ele_is_cpe[i];
    tag = elp->ele_tag[i];
    if (enter_chn(id, tag, 'l', cw, mip) < 0)
      return (-1);
  }

  /* coupling channel elements */
  elp = &pcp->coupling;
  for (i = 0, j = elp->num_ele; i < j; i++)
  {
    mip->cc_tag[i] = elp->ele_tag[i];
    mip->cc_ind[i] = elp->ele_is_cpe[i];
  }
  mip->ncch = j;

  /* mono mixdown elements */
  mxp = &pcp->mono_mix;
  if (mxp->present)
  {
    PRINT(SE,"Unanticipated mono mixdown channel\n");
    return (-1);
  }

  /* stereo mixdown elements */
  mxp = &pcp->stereo_mix;
  if (mxp->present)
  {
    PRINT(SE,"Unanticipated stereo mixdown channel\n");
    return (-1);
  }


  /* matrix mixdown index */
#ifdef AMD4_S

	if(pcp->matrix_mix.present == 1)
	{
		if (pcp->front.num_ele == 2
			&& pcp->front.ele_is_cpe[0] == 0	/* SCE:center */
			&& pcp->front.ele_is_cpe[1] == 1	/* CPE:front L/R */
			&& pcp->side.num_ele == 0
			&&  pcp->back.num_ele == 1
			&& pcp->back.ele_is_cpe[0] == 1		/* CPE:back L/R */
			) 
		{
			;
		}
		else
		{
			/* In a program other than 3/2 configuration, matrix_mixdown_index shall not be contained. */
			printf("In a program other than 3/2 configuration, matrix_mixdown_index shall not be contained\n");	
			return -1;
		}
	}
#else /* AMD4_S */
  mxp = &pcp->matrix_mix;
  if (mxp->present)
  {
    /* removed error return: matrix mixdown index present does not break default decoding */
    /* PRINT(SE,"matrix mixdown index\n"); */
    /* return (-1); */
  }
#endif /* AMD4_S */




  /* save to check future consistency */
  check_mc_info(&mc_info, 1);

  return 1;
}

/*
 * get "closest" sampling frequency index for odd sampling rates
 * (ISO/IEC 14496-3 Table 4.55)
 */
int getNextSampFreqIndex(int sampRate)
{
  int i = 0;
  if (sampRate < sampling_boundarys[(1 << LEN_SAMP_IDX) - 1])
    return -1;
  while (sampRate < sampling_boundarys[i])
  {
    i++;
  }
  return i;
}

/*
 * profile dependent parameters
 */
int tns_max_bands(int islong)
{
  int i;

  /* construct second index */
  i = islong ? 0 : 1;
  i += (mc_info.profile == SSR_Profile) ? 2 : 0;

  return tns_max_bands_tbl[mc_info.sampling_rate_idx][i];
}

int tns_max_order(int islong)
{
  if (islong)
  {
    switch (mc_info.profile)
    {
      case Main_Profile:
#ifdef MPEG4V1
      case LTP_Profile:
#endif
        return 20;
      case LC_Profile:
      case SSR_Profile:
        return 12;
    }
  }
  else
  {
    /* short window */
    return 7;
  }
  return 0;
}

int pred_max_bands(void)
{
  return pred_max_bands_tbl[mc_info.sampling_rate_idx];
}

#define LF	"%-25s %d\n"
#define LFS	"%-25s %d (%s)\n"

/*
 * program configuration element
 */
static void get_ele_list(EleList *p, int enable_cpe)
{
  int i, j;
  for (i = 0, j = p->num_ele; i < j; i++)
  {
    if (enable_cpe)
      p->ele_is_cpe[i] = getbits(LEN_ELE_IS_CPE);
    else
      p->ele_is_cpe[i] = 0; /* sdb */
    p->ele_tag[i] = getbits(LEN_TAG);
  }
}


#ifdef AMD4_S

static unsigned int get_data(
                             unsigned char **pp_stream,
                             unsigned int *p_word,
                             unsigned int *p_vbits,
                             unsigned int *t_bits,
                             int    nbits)
{
  if (*p_vbits <= 16) {
    *p_vbits += 16;
    *p_word <<= 16;
    *p_word |= (unsigned int)*((*pp_stream)++) << 8;
    *p_word |= (unsigned int)*((*pp_stream)++);
  }

  *p_vbits -= nbits;
  *t_bits += nbits;

  return ((*p_word >> *p_vbits) & ((1L << nbits) - 1));
}

static byte getbytefrom_stream(
                               const byte *p_stream,
                               int   nbytes_given,
                               int   curr)
{
  if (curr < nbytes_given) {
    return p_stream[curr];
  } else {
    return 0;
  }
}

static int calculate_crc8_height(
                                 const byte *p_stream,
                                 int         nbytes_given,
                                 crcinfo    *p_crc,
                                 int         check_data
                                 )
{
  register unsigned char crc = 0xFF;
  register byte input;
  int bgn, size;
  int  bgn_offs;
  int  diff;
  int  j;
  int  rem_diff = 0;

  bgn = p_crc->a_elem[0].bgn >> 3;
  diff = p_crc->a_elem[0].end - p_crc->a_elem[0].bgn;
  size = diff >> 3;
  bgn_offs = p_crc->a_elem[0].bgn & 0x7;
  rem_diff = diff & 0x7;

  for (j = bgn; j < bgn + size; j++) {
    input = (byte)(getbytefrom_stream(p_stream, nbytes_given, j) << bgn_offs);
    input |= (byte)(getbytefrom_stream(p_stream, nbytes_given, j+1) >> (8 - bgn_offs));
    crc = crc8_tbl[crc ^ input];
  }

  if (rem_diff) {
    input = (byte)(getbytefrom_stream(p_stream, nbytes_given, j) << bgn_offs);
    input |= (byte)(getbytefrom_stream(p_stream, nbytes_given, j+1) >> (8 - bgn_offs));
    input &= (0xff << (8 - rem_diff));
    while (rem_diff--) {
      register unsigned int msb;
      msb = (input >> 7) ^ (crc >> 7);
      crc <<= 1;
      input <<= 1;
      if (msb & 1) {
        crc ^= 0x07;
      }
    }
  }

  if ((unsigned short)check_data != crc) {
    return -1;
  }

  return 1;
}

/**********************************************************
 Function for parsing Element Height Information in PCE
***********************************************************/
static void height_extension_element(ProgConfig *pce)
{     
  unsigned char *p_stream = (unsigned char*) pce->comments;
  unsigned char *p_streamCRC = (unsigned char*) pce->comments;
  unsigned char **pp_stream = &p_stream;
  unsigned int word = 0;
  unsigned int vbits = 0;
  unsigned int tbits = 0;
  unsigned int syncword;
  int    height_info_crc_check;
  crcinfo crcHeightInfo;
  int    ret;
  int    HeightInfoOKFlag = -1;

  syncword = get_data(pp_stream, &word, &vbits, &tbits, 8); /* syncword */
  while(tbits < (unsigned int) pce->comment_field_bytes*8){
    if (syncword == 0xac ) {
      int j,cLen = 0;
      crcHeightInfo.ele_cnt = 0;
      crcHeightInfo.a_elem[crcHeightInfo.ele_cnt].bgn = tbits-8; /* count syncword for crc */

      for (j=0;j<pce->front.num_ele;j++) {
        pce->a_front_element_height_info[j] = get_data(pp_stream, &word, &vbits, &tbits, 2);
        cLen+=2;
      }
      for (j=0;j<pce->side.num_ele;j++) {
        pce->a_side_element_height_info[j] = get_data(pp_stream, &word, &vbits, &tbits, 2);
        cLen+=2;
      }
      for (j=0;j<pce->back.num_ele;j++) {
        pce->a_back_element_height_info[j] = get_data(pp_stream, &word, &vbits, &tbits, 2);
        cLen+=2;
      }

      /* Byte alignment */
      get_data(pp_stream, &word, &vbits, &tbits, cLen%8);

      crcHeightInfo.a_elem[crcHeightInfo.ele_cnt].end = tbits;
      crcHeightInfo.a_elem[crcHeightInfo.ele_cnt].id = ID_CRC_HEIGHT_INFO;
      crcHeightInfo.ele_cnt = 1;

      height_info_crc_check = get_data(pp_stream, &word, &vbits, &tbits, 8);
      ret = calculate_crc8_height(p_streamCRC, pce->comment_field_bytes, &crcHeightInfo, height_info_crc_check);
      if(ret == 1){
        HeightInfoOKFlag = 1;
        break;
      }
    }
    syncword = (syncword << 1) & 0xFFFF;
    syncword |= get_data(pp_stream, &word, &vbits, &tbits, 1);
  }
}

#endif /* AMD4_S */

#ifndef NO_CORRIGENDUM6
int get_prog_config(ProgConfig *p, int resetMcInfo)
#else
int get_prog_config(ProgConfig *p)
#endif
{
  int i, j, tag;

#ifdef AMD4_S 
 ProgConfig *pcp = p;
 MC_Info *mip = &mc_info;
 int chindex=0;
#endif /* AMD4_S */

  tag = getbits(LEN_TAG);

  p->profile = getbits(LEN_PROFILE);
  p->sampling_rate_idx = getbits(LEN_SAMP_IDX);
  p->front.num_ele = getbits(LEN_NUM_ELE);
  p->side.num_ele = getbits(LEN_NUM_ELE);
  p->back.num_ele = getbits(LEN_NUM_ELE);
  p->lfe.num_ele = getbits(LEN_NUM_LFE);
  p->data.num_ele = getbits(LEN_NUM_DAT);
  p->coupling.num_ele = getbits(LEN_NUM_CCE);
  if ((p->mono_mix.present = getbits(LEN_MIX_PRES)) == 1)
    p->mono_mix.ele_tag = getbits(LEN_TAG);
  if ((p->stereo_mix.present = getbits(LEN_MIX_PRES)) == 1)
    p->stereo_mix.ele_tag = getbits(LEN_TAG);
  if ((p->matrix_mix.present = getbits(LEN_MIX_PRES)) == 1)
  {
    p->matrix_mix.ele_tag = getbits(LEN_MMIX_IDX);
    p->matrix_mix.pseudo_enab = getbits(LEN_PSUR_ENAB);
  }
  get_ele_list(&p->front, 1);
  get_ele_list(&p->side, 1);
  get_ele_list(&p->back, 1);
  get_ele_list(&p->lfe, 0);
  get_ele_list(&p->data, 0);
  get_ele_list(&p->coupling, 1);

  byte_align();
  j = getbits(LEN_COMMENT_BYTES);
  for (i = 0; i < j; i++)
    p->comments[i] = getbits(LEN_BYTE);
  p->comments[i] = 0; /* null terminator for string */

#ifdef AMD4_S 
 p->comment_field_bytes = j;
 /* Element Height Information is extraced from Comment Fields */
 height_extension_element(p);
 
 for (j=0;j<pcp->front.num_ele;j++) 
 {
  mip->front_element_height_info[j] = pcp->a_front_element_height_info[j] ;

  if(pcp->front.ele_is_cpe[j] == 1)
  {
   mip->front_height_info[chindex].cpe = 1;
   mip->front_height_info[chindex + 1 ].cpe = 1;
   if(mip->front_element_height_info[j] == 1)
   {
    mip->front_height_info[chindex++].value = 1;
    mip->front_height_info[chindex++].value = 1;
   }
   else if(mip->front_element_height_info[j] == 2)
   {
    mip->front_height_info[chindex++].value = 2;
    mip->front_height_info[chindex++].value = 2;
   }
   else
   {
    mip->front_height_info[chindex++].value = 0;
    mip->front_height_info[chindex++].value = 0;
   }
  }
  else
  {
   mip->front_height_info[chindex].cpe = 0;
   if(mip->front_element_height_info[j] == 1)
   {
    mip->front_height_info[chindex++].value = 1;    
   }
   else if(mip->front_element_height_info[j] == 2)
   {
    mip->front_height_info[chindex++].value = 2;    
   }
   else
   {
    mip->front_height_info[chindex++].value = 0;    
   }
  }
 }

 chindex = 0;
 for (j=0;j<pcp->side.num_ele;j++) 
 {
  mip->side_element_height_info[j] = pcp->a_side_element_height_info[j] ; 

  if(pcp->side.ele_is_cpe[j] == 1)
  {
   mip->side_height_info[chindex].cpe = 1;
   mip->side_height_info[chindex + 1 ].cpe = 1;
   if(mip->side_element_height_info[j] == 1)
   {
    mip->side_height_info[chindex++].value = 1;
    mip->side_height_info[chindex++].value = 1;
   }
   else if(mip->side_element_height_info[j] == 2)
   {
    mip->side_height_info[chindex++].value = 2;
    mip->side_height_info[chindex++].value = 2;
   }
   else
   {
    mip->side_height_info[chindex++].value = 0;
    mip->side_height_info[chindex++].value = 0;
   }
  }
  else
  {
   mip->side_height_info[chindex].cpe = 0;
   if(mip->side_element_height_info[j] == 1)
   {
    mip->side_height_info[chindex++].value = 1;    
   }
   else if(mip->side_element_height_info[j] == 2)
   {
    mip->side_height_info[chindex++].value = 2;    
   }
   else
   {
    mip->side_height_info[chindex++].value = 0;    
   }
  }
 }
 chindex = 0;
 for (j=0;j<pcp->back.num_ele;j++) 
 {
  mip->back_element_height_info[j] = pcp->a_back_element_height_info[j] ; 

  if(pcp->back.ele_is_cpe[j] == 1)
  {
   mip->back_height_info[chindex].cpe = 1;
   mip->back_height_info[chindex + 1 ].cpe = 1;
   if(mip->back_element_height_info[j] == 1)
   {
    mip->back_height_info[chindex++].value = 1;
    mip->back_height_info[chindex++].value = 1;
   }
   else if(mip->back_element_height_info[j] == 2)
   {
    mip->back_height_info[chindex++].value = 2;
    mip->back_height_info[chindex++].value = 2;
   }
   else
   {
    mip->back_height_info[chindex++].value = 0;
    mip->back_height_info[chindex++].value = 0;
   }
  }
  else
  {
   mip->back_height_info[chindex].cpe = 0;
   if(mip->back_element_height_info[j] == 1)
   {
    mip->back_height_info[chindex++].value = 1;    
   }
   if(mip->back_element_height_info[j] == 2)
   {
    mip->back_height_info[chindex++].value = 2;    
   }
   else
   {
    mip->back_height_info[chindex++].value = 0;    
   }
  }
 } 
#endif /* AMD4_S */



  /* activate new program configuration if appropriate */
  if (current_program < 0)
    current_program = tag; /* select first program found */
#ifndef NO_CORRIGENDUM6
  if (resetMcInfo) {
#endif
  if (tag == current_program)
  {
    /* enter configuration into MC_Info structure */
    if (enter_mc_info(&mc_info, p) < 0)
      return -1;
    /* inhibit default configuration */
    default_config = 0;
  }
#ifndef NO_CORRIGENDUM6
  }
#endif

  return tag;
}

static void print_ele(char *s, EleList *p)
{
  int i;
  PRINT(SE, LF, s, p->num_ele);
  if (p->num_ele)
  {
    if (!strstr(s, "coupling"))
      PRINT(SE, "%6s %6s %6s\n", "idx", "cpe", "tag");
    else
      PRINT(SE, "%6s %6s %6s\n", "idx", "ind", "tag");
  }
  for (i = 0; i < p->num_ele; i++)
    PRINT(SE, "%6d %6d %6d\n", i, p->ele_is_cpe[i], p->ele_tag[i]);
}

void print_adif_header(ADIF_Header *p)
{
  PRINT(SE, "\nADIF header:\n");
  PRINT(SE, LF, "p->copy_id_present", p->copy_id_present);
  PRINT(SE, LF, "p->original_copy", p->original_copy);
  PRINT(SE, LF, "p->home", p->home);
  PRINT(SE, LF, "p->bitstream_type", p->bitstream_type);
  PRINT(SE, LF, "p->bitrate", (int) p->bitrate);
  PRINT(SE, LF, "p->buffer_fullness", (int)p->buffer_fullness );
}

void print_adts_header(ADTS_Header *p)
{
  PRINT(SE, "\nADTS header:\n");
  PRINT(SE, LF, "p->id", (int)p->id );
  PRINT(SE, LF, "p->layer", (int)p->layer );
  PRINT(SE, LF, "p->profile", (int)p->profile );
  PRINT(SE, LF, "p->sampling_rate", (int)sampling_rate[p->sampling_freq_idx] );
  PRINT(SE, LF, "p->buffer_fullness", (int)(p->adts_buffer_fullness<<5) );
  PRINT(SE, LF, "p->channel_config", (int)p->channel_config );
  PRINT(SE, LF, "p->protection_absent", (int)p->protection_abs );
  PRINT(SE, LF, "p->num_of_rdb", (int)p->num_of_rdb );
}

void print_PCE(ProgConfig *p)
{
  PRINT(SE, LFS, "p->profile", p->profile, LOOKUPSTRING(profile, p->profile));
  PRINT(SE, LFS, "p->sampling_rate_idx", p->sampling_rate_idx, LOOKUPSTRING(sfi, p->sampling_rate_idx));

  print_ele("p->front", &p->front);
  print_ele("p->side", &p->side);
  print_ele("p->back", &p->back);
  print_ele("p->lfe", &p->lfe);
  print_ele("p->coupling", &p->coupling);

  PRINT(SE, LF, "p->data.num_ele", p->data.num_ele);
  PRINT(SE, LF, "p->mono_mix.present", p->mono_mix.present);
  PRINT(SE, LF, "p->stereo_mix.present", p->stereo_mix.present);
  PRINT(SE, LF, "p->matrix_mix.present", p->matrix_mix.present);
  PRINT(SE, "%-25s %s\n", "p->comment", p->comments);
}

  /* translate prog config or default config
   * into to multi-channel info structure
   *  returns index of channel in MC_Info
   */
int enter_chn(int id, int tag, char position, int common_window, MC_Info *mip)
{
  int nch, cidx = 0;
  Ch_Info *cip;

  nch = (id == ID_CPE) ? 2 : 1;

  switch (position)
  {
    /* use configuration already in MC_Info, but now common_window is valid */
    case 0:
      cidx = ch_index(mip, id, tag);

      if (nch == 1)
      {
        mip->ch_info[cidx].widx = cidx; /* only one channel for window info*/
      }
      else
      {
        if (common_window)
        {
          mip->ch_info[cidx].widx = cidx; /* window info is left */
          mip->ch_info[cidx + 1].widx = cidx;
        }
        else
        {
          mip->ch_info[cidx].widx = cidx; /* each has window info */
          mip->ch_info[cidx + 1].widx = cidx + 1;
        }
      }
      return cidx;

      /* build configuration */
    case 'f':
      if (mip->nfch + nch > FChans)
      {
        PRINT(SE,"Unanticipated front channel\n");
        return -1;
      }
      if (mip->nfch == 0)
      {
        /* consider case of center channel */
        if (FCenter)
        {
          if (id == ID_CPE)
          {
            /* has center speaker but center channel missing */
            cidx = 0 + 1;
            mip->nfch = 1 + nch;
          }
          else
          {
            if (mip->nfsce & 1)
            {
              /* has center speaker and this is center channel */
              /* odd number of leading SCE's */
              cidx = 0;
              mip->nfch = nch;
            }
            else
            {
              /* has center speaker but center channel missing */
              /* even number of leading SCE's */
              /* (Note that in implicit congiguration
               * channel to speaker mapping may be wrong
               * for first block while count of SCE's prior
               * to first CPE is being make. However first block
               * is not written so it doesn't matter.
               * Second block will be correct.
               */
              cidx = 0 + 1;
              mip->nfch = 1 + nch;
            }
          }
        }
        else
        {
          if (id == ID_CPE)
          {
            /* no center speaker and center channel missing */
            cidx = 0;
            mip->nfch = nch;
          }
          else
          {
            /* no center speaker so this is left channel */
            cidx = 0;
            mip->nfch = nch;
          }
        }
      }
      else
      {
        cidx = mip->nfch;
        mip->nfch += nch;
      }
      break;
    case 's':
      if (mip->nsch + nch > SChans)
      {
        PRINT(SE,"Unanticipated side channel\n");
        return -1;
      }
      cidx = FChans + mip->nsch;
      mip->nsch += nch;
      break;
    case 'b':
      if (mip->nbch + nch > BChans)
      {
        PRINT(SE,"Unanticipated back channel\n");
        return -1;
      }
      cidx = FChans + SChans + mip->nbch;
      mip->nbch += nch;
      break;
    case 'l':
      if (mip->nlch + nch > LChans)
      {
        PRINT(SE,"Unanticipated LFE channel\n");
        return -1;
      }
      cidx = FChans + SChans + BChans + mip->nlch;
      mip->nlch += nch;
      break;
    default:
      myexit("enter_chn");
  }
  mip->nch += nch;

  if (id != ID_CPE)
  {
    /* SCE */
    cip = &mip->ch_info[cidx];
    cip->present = 1;
    cip->tag = tag;
    cip->cpe = 0;
    cip->common_window = common_window;
    cip->widx = cidx;
    /* mip->nch = cidx + 1; */
  }
  else
  {
    /* CPE */
    /* left */
    cip = &mip->ch_info[cidx];
    cip->present = 1;
    cip->tag = tag;
    cip->cpe = 1;
    cip->common_window = common_window;
    cip->ch_is_left = 1;
    cip->paired_ch = cidx + 1;
    /* right */
    cip = &mip->ch_info[cidx + 1];
    cip->present = 1;
    cip->tag = tag;
    cip->cpe = 1;
    cip->common_window = common_window;
    cip->ch_is_left = 0;
    cip->paired_ch = cidx;
    if (common_window)
    {
      mip->ch_info[cidx].widx = cidx; /* window info is left */
      mip->ch_info[cidx + 1].widx = cidx;
    }
    else
    {
      mip->ch_info[cidx].widx = cidx; /* each has window info */
      mip->ch_info[cidx + 1].widx = cidx + 1;
    }
    /* mip->nch = cidx + 2; */
  }

  return cidx;
}

#ifndef AMD4
static char default_position(MC_Info *mip, int id, int* first_cpe)
#else
static char default_position(MC_Info *mip, int id, int* first_cpe, int channelConfigurationValue)
#endif
{
#ifdef AMD4
  if (mip->nch==0) {
    *first_cpe = 0;
  }

  if (channelConfigurationValue > 3) {
    if ((mip->nch == 3 && channelConfigurationValue == 7) ||
      (mip->nch == 3 && channelConfigurationValue == 13) ||
      (mip->nch == 6 && channelConfigurationValue == 14) ||
      (mip->nch == 12 && channelConfigurationValue == 13) ||
      (mip->nch == 13 && channelConfigurationValue == 13) ||
      (mip->nch == 21 && channelConfigurationValue == 13) ||
      (mip->nch == 22 && channelConfigurationValue == 13)) {
        if (id == ID_CPE)
          *first_cpe = 1;
        else if (!(*first_cpe))
          mip->nfsce++;
        return 'f';
    }
    else if ((mip->nch == 5 && channelConfigurationValue == 13) ||
      (mip->nch == 15 && channelConfigurationValue == 13) ||
      (mip->nch == 17 && channelConfigurationValue == 13)) {
        return 's';
    }
    else if ((mip->nch == 3 && channelConfigurationValue == 4) ||
      (mip->nch == 3 && channelConfigurationValue == 5) ||
      (mip->nch == 3 && channelConfigurationValue == 6) ||
      (mip->nch == 3 && channelConfigurationValue == 11) ||
      (mip->nch == 3 && channelConfigurationValue == 12) ||
      (mip->nch == 3 && channelConfigurationValue == 14) ||
      (mip->nch == 5 && channelConfigurationValue == 7) ||
      (mip->nch == 5 && channelConfigurationValue == 11) ||
      (mip->nch == 5 && channelConfigurationValue == 12) ||
      (mip->nch == 7 && channelConfigurationValue == 13) ||
      (mip->nch == 9 && channelConfigurationValue == 13) ||
      (mip->nch == 18 && channelConfigurationValue == 13) ||
      (mip->nch == 20 && channelConfigurationValue == 13)) {
        return 'b';
    }
  }
#endif

  if (mip->nch < FChans)
  {
    if (id == ID_CPE)
      /* here is the first CPE */
      *first_cpe = 1;
    else if ((bno == 0) && !(*first_cpe))
      /* count number of SCE prior to first CPE in first block */
      mip->nfsce++;
    return ('f');
  }
  else if (mip->nch < FChans + SChans)
  {
    return ('s');
  }
  else if (mip->nch < FChans + SChans + BChans)
  {
    return ('b');
  }

  return 0;
}

/* retrieve appropriate channel index for the program
 * and decoder configuration
 */
#ifndef AMD4
int chn_config(int id, int tag, int common_window, MC_Info *mip, int* first_cpe)
#else
int chn_config(int id, int tag, int common_window, MC_Info *mip, int* first_cpe, int channelConfigurationValue)
#endif
{
  int cpe, cidx = 0;
  char position;
#ifdef AMD4
  static int frontElements = 0, sideElements = 0, backElements = 0;
 
  if (mip->nch == 0)
  {
    frontElements = 0, sideElements = 0, backElements = 0;
  }
#endif

  /* channel index to position mapping for 5.1 configuration is
   *  0 center
   *  1 left  front
   *  2 right front
   *  3 left surround
   *  4 right surround
   *  5 lfe
   */

  cpe = (id == ID_CPE) ? 1 : 0;

  if (default_config)
  {
    switch (id)
    {
      case ID_SCE:
      case ID_CPE:
#ifndef AMD4
        if ((position = default_position(mip, id, first_cpe)) == 0)
#else
        if ((position = default_position(mip, id, first_cpe, channelConfigurationValue)) == 0)
#endif
        {
          /*     if ((position = default_position(mip, id)) == 0) { */
          PRINT(SE,"Unanticipated channel\n");
          myexit("default_position");
        }
        cidx = enter_chn(id, tag, position, common_window, mip);
#ifdef AMD4
        if (channelConfigurationValue == 13)
        {
          if (position == 'f')
          {
            if (id == ID_SCE && tag == 2)
            {
              mip->front_element_height_info[frontElements] = 1;
              mip->front_height_info[cidx].value = 1;
              mip->front_height_info[cidx].cpe = 0;
            }
            if (id == ID_CPE && tag == 4)
            {
              mip->front_element_height_info[frontElements] = 1;
              mip->front_height_info[cidx].value = 1;
              mip->front_height_info[cidx+1].value = 1;
              mip->front_height_info[cidx].cpe = 1;
              mip->front_height_info[cidx+1].cpe = 1;
            }
            if (id == ID_SCE && tag == 5)
            {
              mip->front_element_height_info[frontElements] = 2;
              mip->front_height_info[cidx].value = 2;
              mip->front_height_info[cidx].cpe = 0;
            }
            if (id == ID_CPE && tag == 7)
            {
              mip->front_element_height_info[frontElements] = 2;
              mip->front_height_info[cidx].value = 2;
              mip->front_height_info[cidx+1].value = 2;
              mip->front_height_info[cidx].cpe = 1;
              mip->front_height_info[cidx+1].cpe = 1;
            }
          }
          if (position == 's')
          {
            if (id == ID_SCE && tag == 3)
            {
              mip->side_element_height_info[sideElements] = 1;
              mip->side_height_info[cidx-FChans].value = 1;
              mip->side_height_info[cidx-FChans].cpe = 0;
            }
            if (id == ID_CPE && tag == 5)
            {
              mip->side_element_height_info[sideElements] = 1;
              mip->side_height_info[cidx-FChans].value = 1;
              mip->side_height_info[cidx-FChans+1].value = 1;
              mip->side_height_info[cidx-FChans].cpe = 1;
              mip->side_height_info[cidx-FChans+1].cpe = 1;
            }
          }
          if (position == 'b')
          {
            if (id == ID_SCE && tag == 4)
            {
              mip->back_element_height_info[backElements] = 1;
              mip->back_height_info[cidx-FChans-SChans].value = 1;
              mip->back_height_info[cidx-FChans-SChans].cpe = 0;
            }
            if (id == ID_CPE && tag == 6)
            {
              mip->back_element_height_info[backElements] = 1;
              mip->back_height_info[cidx-FChans-SChans].value = 1;
              mip->back_height_info[cidx-FChans-SChans+1].value = 1;
              mip->back_height_info[cidx-FChans-SChans].cpe = 1;
              mip->back_height_info[cidx-FChans-SChans].cpe = 1;
            }
          }
        }
        if (channelConfigurationValue == 14)
        {
          if (position == 'f' && id == ID_CPE && tag == 2)
          {
            mip->front_element_height_info[frontElements] = 1;
            mip->front_height_info[cidx].value = 1;
            mip->front_height_info[cidx+1].value = 1;
            mip->front_height_info[cidx].cpe = 1;
            mip->front_height_info[cidx+1].cpe = 1;
          }
        }
        switch (position)
        {
        case 'f': frontElements++;
          break;
        case 's': sideElements++;
          break;
        case 'b': backElements++;
        }
#endif
        break;
      case ID_LFE:
        cidx = enter_chn(id, tag, 'l', common_window, mip);
        break;
    }
  }
  else
  {
    cidx = enter_chn(id, tag, 0, common_window, mip);
  }

  return cidx; /* index of chn in mc_info */
}

/*
 * check continuity of configuration from one
 * block to next
 */
void reset_mc_info(MC_Info *mip)
{
  int i;
  Ch_Info *p;

  /* always reset number of coupling channels per audio channel */
  for (i = 0; i < Chans; i++)
    mip->ch_info[i].ncch = 0;

  /* only reset for implicit configuration */
  if (default_config)
  {
    /* reset channel counts */
    mip->nch = 0;
    mip->nfch = 0;
    mip->nsch = 0;
    mip->nbch = 0;
    mip->nlch = 0;
    mip->ncch = 0;
    /* reset prior to first block scan only! */
    if (bno == 0)
      mip->nfsce = 0;
    for (i = 0; i < Chans; i++)
    {
      p = &mip->ch_info[i];
      p->present = 0;
      p->cpe = 0;
      p->ch_is_left = 0;
      p->paired_ch = 0;
      p->is_present = 0;
      p->widx = 0;
      p->ncch = 0;
    }
  }
}

void check_mc_info(MC_Info *mip, int new_config)
{
  int i, err;
  Ch_Info *s, *p;
  static MC_Info save_mc_info;

  if (new_config)
  {
    /* enter valid configuration */
    for (i = 0; i < Chans; i++)
    {
      s = &save_mc_info.ch_info[i];
      p = &mip->ch_info[i];
      s->present = p->present;
      s->cpe = p->cpe;
      s->ch_is_left = p->ch_is_left;
      s->paired_ch = p->paired_ch;
    }

    /*count channels for decoderBufferSize*/
    channel_num = mip->nch - mip->nlch;
    for (i = 0; i < mip->ncch; i++)
    {
      channel_num += mip->cc_ind[i];
    }

    /* matejkml LN2 = (mip->frameLengthFlag)? 960 : 1024; */
  }
  else
  {
    /* check this block's configuration */
    err = 0;
    for (i = 0; i < Chans; i++)
    {
      s = &save_mc_info.ch_info[i];
      p = &mip->ch_info[i];
      if (s->present != p->present)
        err = 1;
      if (!s->present)
        continue; /* sdb */
      if (s->cpe != p->cpe)
        err = 1;
      if (s->ch_is_left != p->ch_is_left)
        err = 1;
      if (s->paired_ch != p->paired_ch)
        err = 1;
    }
    if (err)
    {
      PRINT(SE, "Channel configuration inconsistency\n");
      for (i = 0; i < Chans; i++)
      {
        s = &save_mc_info.ch_info[i];
        p = &mip->ch_info[i];
        PRINT(SE, "Channel %d\n", i);
        PRINT(SE, "present    %d %d\n", s->present, p->present);
        PRINT(SE, "cpe        %d %d\n", s->cpe, p->cpe);
        PRINT(SE, "ch_is_left %d %d\n", s->ch_is_left, p->ch_is_left);
        PRINT(SE, "paired_ch  %d %d\n", s->paired_ch, p->paired_ch);
        PRINT(SE, "\n");
      }
      myexit("check_mc_info");
    }
  }
}

void print_mc_info(MC_Info *mip)
{
  int i, j;
  Ch_Info *cip;
  PRINT(SE, "sampling_rate_idx: %i\n", mip->sampling_rate_idx);
  PRINT(SE, "sampling_rate    : %i\n", mip->sampling_rate);
  PRINT(SE, "profile: %i\n", mip->profile);
  PRINT(SE, "Ch Info\n");
  PRINT(SE, "idx tag cpe wdx  cch dom ind\n");
  for (i = 0; i < Chans; i++)
  {
    cip = &mip->ch_info[i];
    if (!cip->present)
      continue;
    PRINT(SE, "%3d %3d %3d %3d ", i, cip->tag, cip->cpe, cip->widx);
    for (j = 0; j < cip->ncch; j++)
    {
      if (j > 0)
        PRINT(SE, "\n%16s", "");
      PRINT(SE, " %3d %3d %3d", cip->cch[j], cip->cc_dom[j], cip->cc_ind[j]);
    }
    PRINT(SE, "\n");
  }
}

/* given id and tag,
 *  returns channel index of SCE or left chn in CPE
 */
int ch_index(MC_Info *mip, int id, int tag)
{
  int ch;
  int cpe = (id == ID_CPE) ? 1 : 0;

  if (id == ID_LFE)
  {
    for (ch = FChans + SChans + BChans; ch < FChans + SChans + BChans + LChans; ch++)
    {

      if (!(mip->ch_info[ch].present))
        continue;

      if (mip->ch_info[ch].tag == tag)
      {
        return ch;
      }
    }
  }
  else
  {
    for (ch = 0; ch < FChans + SChans + BChans; ch++)
    {

      if (!(mip->ch_info[ch].present))
        continue;

      if ((mip->ch_info[ch].cpe == cpe) && (mip->ch_info[ch].tag == tag))
      {
        return ch;
      }
    }
  }

  /* no match, so channel is not in this program
   * dummy up the ch_info structure so rest of chn will parse
   */
  if (XChans > 0)
  {
    ch = Chans - XChans; /* left scratch channel */
    mip->ch_info[ch].cpe = cpe;
    mip->ch_info[ch].ch_is_left = 1;
    mip->ch_info[ch].widx = ch;
    if (cpe)
    {
      mip->ch_info[ch].paired_ch = ch + 1;
      mip->ch_info[ch + 1].ch_is_left = 0;
      mip->ch_info[ch + 1].paired_ch = ch;
    }
  }
  else
  {
    ch = -1; /* error, no scratch space */
  }

  return ch;
}
