/*

This software module was originally developed by

    Shusuke Takahashi (Sony Corporation)

    in the course of development of the MPEG-4 Audio standard (ISO/IEC 14496-3).
    This software module is an implementation of a part of one or more
    MPEG-4 Audio (ISO/IEC 14496-3) tools as specified by the MPEG-4 Audio
    standard (ISO/IEC 14496-3).
    ISO/IEC gives users of the MPEG-4 Audio standards (ISO/IEC 14496-3)
    free license to this software module or modifications thereof for use
    in hardware or software products claiming conformance to the MPEG-4
    Audio standards (ISO/IEC 14496-3).
    Those intending to use this software module in hardware or software
    products are advised that this use may infringe existing patents.
    The original developer of this software module and his/her company,
    the subsequent editors and their companies, and ISO/IEC have no
    liability for use of this software module or modifications thereof in
    an implementation.
    Copyright is not released for non MPEG-4 Audio (ISO/IEC 14496-3)
    conforming products. The original developer retains full right to use
    the code for his/her own purpose, assign or donate the code to a third
    party and to inhibit third party from using the code for non MPEG-4
    Audio (ISO/IEC 14496-3) conforming products.
    This copyright notice must be included in all copies or derivative works.

    Copyright (c)2015.

*/

#ifndef _LIBMPEGASFE_H
#define _LIBMPEGASFE_H

/* data type definition */
typedef unsigned int MPEG_AS_ERRCODE;
typedef unsigned int MPEG_AS_VERSION;
typedef int MPEG_AS_RESULT;
typedef struct _mpeg_as_fe_handle *MPEG_AS_FE_HANDLE;

/* prototype declaration for API functions */
#ifdef __cplusplus
extern "C" {
#endif
#define MPEG_AS_DECLSPEC 

MPEG_AS_DECLSPEC	MPEG_AS_FE_HANDLE	mpeg_as_fe_get_handle(void);
MPEG_AS_DECLSPEC	MPEG_AS_RESULT  mpeg_as_fe_init_handle(MPEG_AS_FE_HANDLE  hData);
MPEG_AS_DECLSPEC	MPEG_AS_RESULT  mpeg_as_fe_free_handle(MPEG_AS_FE_HANDLE  hData);
MPEG_AS_DECLSPEC	MPEG_AS_RESULT  mpeg_as_fe_set_time_resolution(MPEG_AS_FE_HANDLE  hData, int resolution);
MPEG_AS_DECLSPEC	MPEG_AS_RESULT  mpeg_as_fe_get_frame_size(MPEG_AS_FE_HANDLE  hData, int*	p_frame_size);
MPEG_AS_DECLSPEC	MPEG_AS_RESULT  mpeg_as_fe_get_feature_size(MPEG_AS_FE_HANDLE  hData, int*	p_frame_size);
MPEG_AS_DECLSPEC	MPEG_AS_RESULT  mpeg_as_fe_get_sampling_rate(MPEG_AS_FE_HANDLE  hData, int* p_sampling_rate);
MPEG_AS_DECLSPEC	MPEG_AS_RESULT  mpeg_as_fe_processing(MPEG_AS_FE_HANDLE  hData, char* p_inputbuf, char* p_outputbuf, int *p_is_valid);

#ifdef __cplusplus
}
#endif

/* error handling macros */
#define MPEG_AS_S_OK				((MPEG_AS_RESULT)0x00000000L)
#define MPEG_AS_S_FALSE				((MPEG_AS_RESULT)0x00000001L)
#define MPEG_AS_E_FAIL				((MPEG_AS_RESULT)0x80000000L)
#define MPEG_AS_E_BAD_ARGUMENTS		((MPEG_AS_RESULT)0x80000001L)
#define MPEG_AS_E_OUT_OF_MEMORY		((MPEG_AS_RESULT)0x80000002L)
#define MPEG_AS_E_INVALID_HANDLE	((MPEG_AS_RESULT)0x80000003L)
#define MPEG_AS_E_NOT_INITIALIZED	((MPEG_AS_RESULT)0x80000004L)
#define	MPEG_AS_E_SHORT_INPUT		((MPEG_AS_RESULT)0x80000005L)


#define MPEG_AS_SUCCEEDED(result)	((MPEG_AS_RESULT)(result)>=0)
#define MPEG_AS_FAILED(result)	((MPEG_AS_RESULT)(result)<0)

#endif	/* _LIBMPEG_AS_H */
/*******************************************************************************
	End of file
*******************************************************************************/

