/*

This software module was originally developed by

    Shusuke Takahashi (Sony Corporation)

    in the course of development of the MPEG-4 Audio standard (ISO/IEC 14496-3).
    This software module is an implementation of a part of one or more
    MPEG-4 Audio (ISO/IEC 14496-3) tools as specified by the MPEG-4 Audio
    standard (ISO/IEC 14496-3).
    ISO/IEC gives users of the MPEG-4 Audio standards (ISO/IEC 14496-3)
    free license to this software module or modifications thereof for use
    in hardware or software products claiming conformance to the MPEG-4
    Audio standards (ISO/IEC 14496-3).
    Those intending to use this software module in hardware or software
    products are advised that this use may infringe existing patents.
    The original developer of this software module and his/her company,
    the subsequent editors and their companies, and ISO/IEC have no
    liability for use of this software module or modifications thereof in
    an implementation.
    Copyright is not released for non MPEG-4 Audio (ISO/IEC 14496-3)
    conforming products. The original developer retains full right to use
    the code for his/her own purpose, assign or donate the code to a third
    party and to inhibit third party from using the code for non MPEG-4
    Audio (ISO/IEC 14496-3) conforming products.
    This copyright notice must be included in all copies or derivative works.

    Copyright (c)2015.

*/

#include <stdio.h>
#include <stdlib.h>

#include "as_feature.h"

/* prototype */
extern const SCALAR g_a_bpf_real[MPEG_AS_NUM_BAND][MPEG_AS_FFT_LEN];
extern const SCALAR g_a_bpf_imag[MPEG_AS_NUM_BAND][MPEG_AS_FFT_LEN];

static void peak_detection(SCALAR *p_feat, int *p_is_peak);
static void frame_rate_conversion(MPEG_AS_FE_HANDLE hData, int *p_is_peak, int *p_is_valid);

static void pre_emphasis(
	SCALAR *p_data, 
	SCALAR *p_prev_input,
	int	nsample
) 
{
	int		i;
	SCALAR	prev_val, val;
	
	prev_val = *p_prev_input;

	for (i = 0; i < nsample; i++) {
		val = p_data[i];
		p_data[i] = val - (SCALAR)0.97*prev_val;
		prev_val = val;
	}
	
	*p_prev_input = prev_val;

}

/* functions */
int feature_detection (
MPEG_AS_FE_HANDLE hData, 
const short *p_input, 
int *p_is_peak,
int	*p_is_valid
) 
{
	int idx_time;
	int idx_spec;
	SCALAR a_work[MPEG_AS_FFT_LEN];

	SCALAR	a_work2[MPEG_AS_FFT_LEN];
	SCALAR	a_work3[MPEG_AS_FFT_LEN];
	int idx_bpf;

	int num_pitch_band;
	SCALAR a_buf_work[MPEG_AS_FFT_LEN];
	SCALAR a_buf_work_im[MPEG_AS_FFT_LEN];
	SCALAR a_buf_work2[MPEG_AS_FFT_LEN];
	SCALAR a_buf_work2_im[MPEG_AS_FFT_LEN];

	SCALAR normcoef;

	if (!hData->is_initialized) {
		return MPEG_AS_E_NOT_INITIALIZED;
	}
	/* copy previous input signal to local buffer */
	for (idx_time = 0; idx_time < MPEG_AS_WINDOW_LEN - hData->feature_frame_step; idx_time++) {
		a_work[idx_time] = hData->a_signal_buf[idx_time];
	}

	/* copy input signal from application to local buffer */
	for (idx_time = 0; idx_time < hData->feature_frame_step; idx_time++) {
		a_work[MPEG_AS_WINDOW_LEN - hData->feature_frame_step + idx_time] = (SCALAR)p_input[idx_time];
	}

	/* pre-emphasis */
	pre_emphasis(&a_work[MPEG_AS_WINDOW_LEN - hData->feature_frame_step], &(hData->p_prev_input),  hData->feature_frame_step);

	/* zero padding */
	for (idx_time = MPEG_AS_WINDOW_LEN; idx_time < MPEG_AS_FFT_LEN; idx_time++) {
		a_work[idx_time] = 0; 
	}

	for (idx_time = 0; idx_time < MPEG_AS_WINDOW_LEN -  hData->feature_frame_step; idx_time++) {
		hData->a_signal_buf[idx_time] = a_work[idx_time + hData->feature_frame_step];
	}

	/* calculate auto-correlation by using FFT */

	/* ------------------------------------------------*
	 *  FFT computation
	 * ------------------------------------------------*/
	for (idx_time = 0; idx_time < MPEG_AS_WINDOW_LEN; idx_time++) {
		SCALAR win = (SCALAR)(0.54  - 0.46 * cos(2*M_PI*idx_time/(MPEG_AS_WINDOW_LEN-1)));
		a_buf_work[idx_time] = a_work[idx_time]*win;
		a_buf_work_im[idx_time] = 0.0;
	}

	for (idx_time = MPEG_AS_WINDOW_LEN; idx_time < MPEG_AS_FFT_LEN; idx_time++) {
		a_buf_work[idx_time] = 0.0;
		a_buf_work_im[idx_time] = 0.0;
	}

	IPC_fft(a_buf_work, a_buf_work_im, MPEG_AS_FFT_STAGE);


	
	for (idx_spec = 0; idx_spec < MPEG_AS_FFT_LEN; idx_spec++) {
		a_work3[idx_spec] = 0.0;
	}

	num_pitch_band = 0;

	for (idx_bpf = 0; idx_bpf < MPEG_AS_NUM_BAND; idx_bpf++) {
		int startpos = MPEG_AS_MAX_SEARCH_START_POS;
		int endpos = MPEG_AS_MAX_SEARCH_END_POS;

		SCALAR maxval;

		/* ------------------------------------------------*
		 * apply bandpass filter 
		 * ------------------------------------------------*/

		for (idx_spec = 0; idx_spec < hData->npoint; idx_spec++) {
			SCALAR re, im;
			re = g_a_bpf_real[idx_bpf][idx_spec]*a_buf_work[idx_spec] - g_a_bpf_imag[idx_bpf][idx_spec]*a_buf_work_im[idx_spec];
			im = g_a_bpf_real[idx_bpf][idx_spec]*a_buf_work_im[idx_spec] + g_a_bpf_imag[idx_bpf][idx_spec]*a_buf_work[idx_spec];
			
			a_work2[idx_spec] = re*re + im*im;
		}

		/* ------------------------------------------------*
		 * IFFT computation
		 * ------------------------------------------------*/

		for (idx_time = 0; idx_time <= MPEG_AS_FFT_LEN/2; idx_time++) {
			a_buf_work2[idx_time] = (SCALAR)a_work2[idx_time];
			a_buf_work2_im[idx_time] = (SCALAR)0.0;
		}

		for (; idx_time < MPEG_AS_FFT_LEN; idx_time++) {
			a_buf_work2[idx_time] = (SCALAR)a_work2[MPEG_AS_FFT_LEN - idx_time];
			a_buf_work2_im[idx_time] = (SCALAR)0.0;
		}

		IPC_ifft(a_buf_work2, a_buf_work2_im, MPEG_AS_FFT_STAGE);

		for (idx_spec = 0; idx_spec < MPEG_AS_FFT_LEN; idx_spec++) {
			a_work2[idx_spec] = a_buf_work2[idx_spec];
		}

		/* ------------------------------------------------*
		 * calculate normalize coefficient
		 * ------------------------------------------------*/
		if (a_work2[0] == 0.0) {
			normcoef = 0;
		} else {
			normcoef = 1/a_work2[0];
		}

		/* ------------------------------------------------*
		 * calculate pitch gain
		 * ------------------------------------------------*/
		maxval = a_work2[startpos];

		for (idx_spec = startpos + 1; idx_spec < endpos; idx_spec++) {
			if (a_work2[idx_spec] > maxval) {
				maxval = a_work2[idx_spec];
			}
		}

		/* ------------------------------------------------*
		 * apply pitch filter
		 * ------------------------------------------------*/
		if (maxval*normcoef > 0.3) {
			/* accumulate features */
			for (idx_spec = 0; idx_spec < MPEG_AS_FFT_LEN; idx_spec++) {
				a_work3[idx_spec] += a_work2[idx_spec]*normcoef;
			}
			num_pitch_band++;

		}

	}

	if (num_pitch_band != 0) {
		for (idx_spec = 0; idx_spec < MPEG_AS_FFT_LEN; idx_spec++) {
			a_work3[idx_spec] *= (SCALAR)MPEG_AS_NUM_BAND / num_pitch_band;
		}
	}

	/* ------------------------------------------------*
	 * detect peak of features
	 * ------------------------------------------------*/
	peak_detection(a_work3, p_is_peak);

	/* ------------------------------------------------*
	 * frame rate conversion
	 * ------------------------------------------------*/
	frame_rate_conversion(hData, p_is_peak, p_is_valid);

	hData->idx_frame++;

	return MPEG_AS_S_OK;	
}

/* sub functions */

static void peak_detection(SCALAR *p_feat, int *p_is_peak) {
	int ntap_ma = MPEG_AS_NUM_TAP_MA;
	SCALAR *p_powspec = p_feat;
	int num_spec = MPEG_AS_MAX_LAG;

	int idx_spec;

	SCALAR acc_spec = 0.0;

	SCALAR oldest;

	SCALAR a_thresh[MPEG_AS_MAX_LAG];

	SCALAR prevval;

	for (idx_spec = 0; idx_spec < num_spec; idx_spec++) {
		p_is_peak[idx_spec] = 0;
	}

	/* calculate moving average */
	for (idx_spec = 0; idx_spec <= ntap_ma/2; idx_spec++) {
		acc_spec += p_powspec[idx_spec];
	}

	for (idx_spec = 0; idx_spec <= ntap_ma/2; idx_spec++) {
		acc_spec += p_powspec[idx_spec + ntap_ma/2];
		a_thresh[idx_spec] = acc_spec/(idx_spec + 1 + ntap_ma/2);
	}

	oldest = p_powspec[0];

	for (; idx_spec < num_spec-ntap_ma/2; idx_spec++) {
		acc_spec = acc_spec - oldest + p_powspec[idx_spec + ntap_ma/2];
		oldest = p_powspec[idx_spec - ntap_ma/2];
		a_thresh[idx_spec] = acc_spec/(ntap_ma + 1);
	}

	for (; idx_spec < num_spec; idx_spec++) {
		acc_spec = acc_spec - oldest;
		oldest = p_powspec[idx_spec - ntap_ma/2];
		a_thresh[idx_spec] = acc_spec/(num_spec - idx_spec + ntap_ma/2);
	}

	/* peak detection */
	prevval = p_powspec[0];
	for (idx_spec = 1; idx_spec < num_spec-1; idx_spec++) {
		if (p_powspec[idx_spec] > p_powspec[idx_spec + 1] &&
			p_powspec[idx_spec] >= prevval
			) 
		{

			SCALAR mean_spec = a_thresh[idx_spec];
			SCALAR thresh = (SCALAR)0.1;
			prevval = p_powspec[idx_spec];

			if (p_powspec[idx_spec] > mean_spec + thresh) {
				p_is_peak[idx_spec] = 1;
				p_powspec[idx_spec] = p_powspec[idx_spec] - (mean_spec + thresh);


			} else {
				p_powspec[idx_spec] = 0;
				p_is_peak[idx_spec] = 0;
			}


		} else {
			prevval = p_powspec[idx_spec];
			p_is_peak[idx_spec] = 0;
		}
	}

	p_is_peak[0] = 0;
	p_is_peak[num_spec-1] = 0;

	p_powspec[0] = 0;
	p_powspec[num_spec-1] = 0;

}

static void frame_rate_conversion(MPEG_AS_FE_HANDLE hData, int *p_is_peak, int *p_is_valid) {
	
	int idx_lag;

	for (idx_lag = 0; idx_lag < MPEG_AS_MAX_LAG; idx_lag++) {
		hData->aa_peakpos_buf[hData->idx_peakpos_buf][idx_lag] = p_is_peak[idx_lag];
	}

	hData->idx_peakpos_buf++;

	if (hData->idx_peakpos_buf == hData->resolution / 8) {
		/* conversion */
		for (idx_lag = 0; idx_lag < MPEG_AS_MAX_LAG; idx_lag++) {
			int tmpval = 0;
			int	idx_frame;
			for (idx_frame = 0; idx_frame < hData->idx_peakpos_buf; idx_frame++) {
				tmpval |= hData->aa_peakpos_buf[idx_frame][idx_lag];
			}
			p_is_peak[idx_lag] = tmpval;
		}

		hData->idx_peakpos_buf = 0;
		*p_is_valid = 1;
	} else {
		memset(p_is_peak, 0, sizeof(int)*MPEG_AS_MAX_LAG);
		*p_is_valid = 0;
	}

	return;
}

/*******************************************************************************
	End of file
*******************************************************************************/

