/*

This software module was originally developed by

    Shusuke Takahashi (Sony Corporation)

    in the course of development of the MPEG-4 Audio standard (ISO/IEC 14496-3).
    This software module is an implementation of a part of one or more
    MPEG-4 Audio (ISO/IEC 14496-3) tools as specified by the MPEG-4 Audio
    standard (ISO/IEC 14496-3).
    ISO/IEC gives users of the MPEG-4 Audio standards (ISO/IEC 14496-3)
    free license to this software module or modifications thereof for use
    in hardware or software products claiming conformance to the MPEG-4
    Audio standards (ISO/IEC 14496-3).
    Those intending to use this software module in hardware or software
    products are advised that this use may infringe existing patents.
    The original developer of this software module and his/her company,
    the subsequent editors and their companies, and ISO/IEC have no
    liability for use of this software module or modifications thereof in
    an implementation.
    Copyright is not released for non MPEG-4 Audio (ISO/IEC 14496-3)
    conforming products. The original developer retains full right to use
    the code for his/her own purpose, assign or donate the code to a third
    party and to inhibit third party from using the code for non MPEG-4
    Audio (ISO/IEC 14496-3) conforming products.
    This copyright notice must be included in all copies or derivative works.

    Copyright (c)2015.

*/
#ifndef	_AS_FEATURE_H
#define	_AS_FEATURE_H


#include <string.h>

#include "libmpegasfe.h"

/* macro definition */
typedef float SCALAR;
#define _scalar(x) x##f
#define _USE_MATH_DEFINES
#include <math.h>
#ifndef M_PI
#define   M_PI       3.14159265358979323846
#endif


#define	MPEG_AS_FEATURE_FRAME_STEP 64

#define	MPEG_AS_NUM_BYTE_FEAT	16
#define	MPEG_AS_SAMPLING_RATE	8000
#define	MPEG_AS_DEFAULT_RESOLUTION	8

#define	MPEG_AS_NUM_BAND	5

/* FFT */
#define	MPEG_AS_FFT_LEN		1024
#define	MPEG_AS_FFT_STAGE	10

#define	MPEG_AS_WINDOW_LEN		256
#define	MPEG_AS_RATE_CONVERT_BUF_SIZE	4
#define	MPEG_AS_NUM_TAP_MA		20

#define	MPEG_AS_MAX_LAG	128

#define	MPEG_AS_FEATURE_RESOLUTION_8_MSEC	8
#define	MPEG_AS_FEATURE_RESOLUTION_32_MSEC	32

#define	MPEG_AS_MAX_SEARCH_START_POS	10
#define	MPEG_AS_MAX_SEARCH_END_POS		128

/* data type definition */
typedef	struct _mpeg_as_fe_handle {
	int	is_initialized;

	SCALAR	a_signal_buf[MPEG_AS_FFT_LEN];
	SCALAR	p_prev_input;
	int idx_frame;
	
	/* FFT */
	int	npoint;
	int	feature_frame_step;

	int aa_peakpos_buf[MPEG_AS_RATE_CONVERT_BUF_SIZE][MPEG_AS_FFT_LEN];
	int	idx_peakpos_buf;

	int resolution;
} HANDLE_FEATURE_DETECTION_STRUCT;

/* proto type declaration */

void IPC_fft(SCALAR *x, SCALAR *y, int dim);
void IPC_ifft(SCALAR *x, SCALAR *y, int dim);
void IPC_makeCoef(SCALAR	*coef, int	n);

int feature_detection(MPEG_AS_FE_HANDLE hData, const short *p_input, int *p_is_peak, int *p_is_valid);

void *mpeg_as_malloc(int size);
void *mpeg_as_calloc(int count, int size);
void mpeg_as_free(void *ptr);

#endif	/* _AS_FEATURE_H */

/*******************************************************************************
	End of file
*******************************************************************************/

