/*

This software module was originally developed by

    Masayuki Nishiguchi, Kazuyuki Iijima, and Jun Matsumoto (Sony Corporation)

    and edited by

    Akira Inoue and Shusuke Takahashi (Sony Corporation)

    in the course of development of the MPEG-4 Audio standard (ISO/IEC 14496-3).
    This software module is an implementation of a part of one or more
    MPEG-4 Audio (ISO/IEC 14496-3) tools as specified by the MPEG-4 Audio
    standard (ISO/IEC 14496-3).
    ISO/IEC gives users of the MPEG-4 Audio standards (ISO/IEC 14496-3)
    free license to this software module or modifications thereof for use
    in hardware or software products claiming conformance to the MPEG-4
    Audio standards (ISO/IEC 14496-3).
    Those intending to use this software module in hardware or software
    products are advised that this use may infringe existing patents.
    The original developer of this software module and his/her company,
    the subsequent editors and their companies, and ISO/IEC have no
    liability for use of this software module or modifications thereof in
    an implementation.
    Copyright is not released for non MPEG-4 Audio (ISO/IEC 14496-3)
    conforming products. The original developer retains full right to use
    the code for his/her own purpose, assign or donate the code to a third
    party and to inhibit third party from using the code for non MPEG-4
    Audio (ISO/IEC 14496-3) conforming products.
    This copyright notice must be included in all copies or derivative works.

    Copyright (c)2015.

*/

#include "as_feature.h"

static void btfly(SCALAR c, SCALAR s, SCALAR r, SCALAR i, SCALAR *x, SCALAR *y)
{
    *x = c*r + s*i;
    *y = c*i - s*r;
    return;
}

void IPC_fft(SCALAR *x, SCALAR *y, int dim)
{
    int		i,j,k,m,m1,m2;
    int		p,q;
    SCALAR	px, py;
    SCALAR	th, dth;
    SCALAR	a,b;
    int	    ptr1, ptr2;	

    m = 1 << dim;

	q = m;
    dth = (SCALAR)2.0*(SCALAR)M_PI / (SCALAR)m;

    for(i=1; i<=dim; i++) {
	p = q;
	q /= 2;
	th = 0.0;
	for(j=1; j<=q; j++) {
	    px = (SCALAR)cos(th); py = (SCALAR)sin(th);
	    th += dth;
	    for(k=p; k<=m; k+=p) {
		ptr1 = k - p + j;
		ptr2 = ptr1 + q;
		a = x[ptr1-1] - x[ptr2-1];
		b = y[ptr1-1] - y[ptr2-1];
		x[ptr1-1] = x[ptr1-1] + x[ptr2-1];
		y[ptr1-1] = y[ptr1-1] + y[ptr2-1];
		btfly(px, py, a, b, &x[ptr2-1], &y[ptr2-1]);
	    }
	}
	dth *= 2.0;
    }

    j=1;
    m1= m-1;
    m2= m/2;

    for(i=1; i<=m1; i++) {
	if(i<j) {
	    a = x[j-1];
	    b = y[j-1];
	    x[j-1] = x[i-1];
	    y[j-1] = y[i-1];
	    x[i-1] = a;
	    y[i-1] = b;
	}
	k = m2;
	while(k<j) {
	    j -= k;
	    k /= 2;
	}
	j += k;
    }
}

void IPC_ifft(SCALAR *x, SCALAR *y, int dim)
{
    int		m;
    int		i;

    m = 1 << dim;

    for(i=1; i<=m; i++) 
	y[i-1] = -y[i-1];
    IPC_fft(x, y, dim);
    for(i=1; i<=m; i++) {
	x[i-1] = x[i-1]/(SCALAR)m;
	y[i-1] = -y[i-1]/(SCALAR)m;
    }
}

void IPC_makeCoef(
SCALAR	*coef,
int	n)
{
    int i;
    SCALAR energy;
    SCALAR c4;
    
    energy = 0.;
    for(i=0; i < n ; i++){
	coef[i]=((SCALAR)0.54  - (SCALAR)0.46 * (SCALAR)cos( 2*M_PI*i/(n-1))); 
	energy=energy + coef[i] * coef[i] ;
    }
    for(i=0; i < n ; i++)
	coef[i]=coef[i]/(SCALAR)sqrt(energy);
    
    c4 = 0.;
    for(i=0; i < n ; i++)
	c4 = c4 + (SCALAR)pow((double)coef[i], 4.);
}
/*****************************************************************************
	End of file
*****************************************************************************/




