/*

This software module was originally developed by

    Shusuke Takahashi (Sony Corporation)

    in the course of development of the MPEG-4 Audio standard (ISO/IEC 14496-3).
    This software module is an implementation of a part of one or more
    MPEG-4 Audio (ISO/IEC 14496-3) tools as specified by the MPEG-4 Audio
    standard (ISO/IEC 14496-3).
    ISO/IEC gives users of the MPEG-4 Audio standards (ISO/IEC 14496-3)
    free license to this software module or modifications thereof for use
    in hardware or software products claiming conformance to the MPEG-4
    Audio standards (ISO/IEC 14496-3).
    Those intending to use this software module in hardware or software
    products are advised that this use may infringe existing patents.
    The original developer of this software module and his/her company,
    the subsequent editors and their companies, and ISO/IEC have no
    liability for use of this software module or modifications thereof in
    an implementation.
    Copyright is not released for non MPEG-4 Audio (ISO/IEC 14496-3)
    conforming products. The original developer retains full right to use
    the code for his/her own purpose, assign or donate the code to a third
    party and to inhibit third party from using the code for non MPEG-4
    Audio (ISO/IEC 14496-3) conforming products.
    This copyright notice must be included in all copies or derivative works.

    Copyright (c)2015.

*/

#include	<stdlib.h>

#include	"libmpegasfe.h"
#include	"as_feature.h"

/************************************************************************
	API functions
************************************************************************/

/* get handle */
MPEG_AS_FE_HANDLE	mpeg_as_fe_get_handle(void) 
{
	MPEG_AS_FE_HANDLE hData;

	hData = (MPEG_AS_FE_HANDLE)mpeg_as_calloc(1, sizeof(HANDLE_FEATURE_DETECTION_STRUCT));
	if (hData != NULL) {
		hData->feature_frame_step = MPEG_AS_FEATURE_FRAME_STEP;
		hData->resolution = MPEG_AS_DEFAULT_RESOLUTION;
	}

	return hData;	
}

/* init */
MPEG_AS_RESULT  mpeg_as_fe_init_handle (
MPEG_AS_FE_HANDLE  hData
) 
{
	MPEG_AS_RESULT result = MPEG_AS_S_OK;
	int idx_frame;

	if (hData == NULL) {
		return MPEG_AS_E_INVALID_HANDLE;
	}

	hData->is_initialized = 1;

	hData->idx_frame = 0;

	memset(hData->a_signal_buf, 0, sizeof(int)*MPEG_AS_FFT_LEN);

	/* FFT */
	hData->npoint = MPEG_AS_FFT_LEN;


	hData->idx_peakpos_buf = 0;
	for (idx_frame = 0; idx_frame < MPEG_AS_RATE_CONVERT_BUF_SIZE; idx_frame++) {
		memset(hData->aa_peakpos_buf[idx_frame], 0, sizeof(int)*MPEG_AS_FFT_LEN);
	}


	return result;	
}

/* free */
MPEG_AS_RESULT  mpeg_as_fe_free_handle (
MPEG_AS_FE_HANDLE  hData
) 
{
	MPEG_AS_RESULT result = MPEG_AS_S_OK;
	if (hData != NULL) {
		mpeg_as_free(hData);
	}


	return result;
}

/* set time resolution */
MPEG_AS_RESULT  mpeg_as_fe_set_time_resolution(
MPEG_AS_FE_HANDLE  hData, 
int resolution
)
{
		
	MPEG_AS_RESULT result = MPEG_AS_S_OK;

	if (hData == NULL) {
		return MPEG_AS_E_INVALID_HANDLE;
	} 

	if (resolution != MPEG_AS_FEATURE_RESOLUTION_8_MSEC && resolution != MPEG_AS_FEATURE_RESOLUTION_32_MSEC) {
		return MPEG_AS_E_BAD_ARGUMENTS;
	} 

	hData->resolution = resolution;

	return result;
}


/* get frame size */
MPEG_AS_RESULT  mpeg_as_fe_get_frame_size (
MPEG_AS_FE_HANDLE  hData, 
int*	p_frame_size
) 
{
	MPEG_AS_RESULT result = MPEG_AS_S_OK;

	if (hData == NULL) {
		return MPEG_AS_E_INVALID_HANDLE;
	} 

	if (p_frame_size == NULL) {
		return MPEG_AS_E_BAD_ARGUMENTS;
	} 

	*p_frame_size = hData->feature_frame_step;

	return result;
}

/* get feature size */
MPEG_AS_RESULT  mpeg_as_fe_get_feature_size (
MPEG_AS_FE_HANDLE  hData, 
int*	p_frame_size
) 
{
	if (hData == NULL) {
		return MPEG_AS_E_INVALID_HANDLE;
	} 

	if (p_frame_size == NULL) {
		return MPEG_AS_E_BAD_ARGUMENTS;
	} 

	*p_frame_size = MPEG_AS_NUM_BYTE_FEAT;
	return MPEG_AS_S_OK;
}

/* get sampling rate */
MPEG_AS_RESULT  mpeg_as_fe_get_sampling_rate (
MPEG_AS_FE_HANDLE  hData, 
int* p_sampling_rate
) 
{
	if (hData == NULL) {
		return MPEG_AS_E_INVALID_HANDLE;
	} 

	if (p_sampling_rate == NULL) {
		return MPEG_AS_E_BAD_ARGUMENTS;
	} 

	*p_sampling_rate = MPEG_AS_SAMPLING_RATE;
	return MPEG_AS_S_OK;
}

/* feature extraction */
MPEG_AS_RESULT  mpeg_as_fe_processing (
MPEG_AS_FE_HANDLE  hData, 
char* p_inputbuf, 
char* p_outputbuf,
int	*p_is_valid
)
{
	MPEG_AS_RESULT result;
	int p_is_peak[MPEG_AS_MAX_LAG];
	int	idx_spec;
	int idx_byte;
	int idx_bit;

	if (hData == NULL) {
		return MPEG_AS_E_INVALID_HANDLE;
	}

	if (p_inputbuf == NULL) {
		return MPEG_AS_E_BAD_ARGUMENTS;
	}

	if (p_outputbuf == NULL) {
		return MPEG_AS_E_BAD_ARGUMENTS;
	}

	result = feature_detection (hData, (short *)p_inputbuf, p_is_peak, p_is_valid);

	if (result != MPEG_AS_S_OK) {
		return result;
	}

	/* peak to bit pattern */
	if (*p_is_valid == 1) {
		for (idx_byte = 0; idx_byte < MPEG_AS_NUM_BYTE_FEAT; idx_byte++) {
			char tmp = 0;

			for (idx_bit = 0; idx_bit < 8; idx_bit++) {
				idx_spec = idx_byte*8 + idx_bit;
				tmp = tmp | (char)(p_is_peak[idx_spec] << (7 - idx_bit));
			}
			p_outputbuf[idx_byte] = tmp;
		}
	} else {
		memset(p_outputbuf, 0, MPEG_AS_NUM_BYTE_FEAT);
	}

	return MPEG_AS_S_OK;
}


void *mpeg_as_malloc(int size) {
	void *ptr;

	ptr = malloc(size);

	return ptr;
}

void *mpeg_as_calloc(int count, int size) {
	void *ptr;
			
	ptr = calloc(count, size);

	return ptr;
}

void mpeg_as_free(void *ptr) {
	
	free(ptr);

	return;
}
/*******************************************************************************
	End of file
*******************************************************************************/

