/*

This software module was originally developed by

    Shusuke Takahashi (Sony Corporation)

    in the course of development of the MPEG-4 Audio standard (ISO/IEC 14496-3).
    This software module is an implementation of a part of one or more
    MPEG-4 Audio (ISO/IEC 14496-3) tools as specified by the MPEG-4 Audio
    standard (ISO/IEC 14496-3).
    ISO/IEC gives users of the MPEG-4 Audio standards (ISO/IEC 14496-3)
    free license to this software module or modifications thereof for use
    in hardware or software products claiming conformance to the MPEG-4
    Audio standards (ISO/IEC 14496-3).
    Those intending to use this software module in hardware or software
    products are advised that this use may infringe existing patents.
    The original developer of this software module and his/her company,
    the subsequent editors and their companies, and ISO/IEC have no
    liability for use of this software module or modifications thereof in
    an implementation.
    Copyright is not released for non MPEG-4 Audio (ISO/IEC 14496-3)
    conforming products. The original developer retains full right to use
    the code for his/her own purpose, assign or donate the code to a third
    party and to inhibit third party from using the code for non MPEG-4
    Audio (ISO/IEC 14496-3) conforming products.
    This copyright notice must be included in all copies or derivative works.

    Copyright (c)2015.

*/
#ifndef _WAVEFILE_H
#define _WAVEFILE_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define RIFFWAVE_OK  0
#define RIFFWAVE_NG -1

#define WAVEXT_CHMASK_5_1	(0x003f)
#define WAVEXT_CHMASK_7_1	(0x063f)		

typedef struct {
	union {
		short wValidBitsPerSample;
		short wSamplesPerBlock;
		short wReserved;
	} Samples;
	unsigned int dwChannelMask;
} EXTENSIBLE;

typedef struct {
	int   chunk_size;
	short wFormatTag;
	short nChannels;
	int   nSamplesPerSec;
	int   nAvgBytesPerSec;
	short nBlockAlign;
	short wBitsPerSample;
	EXTENSIBLE ext;
} FMTCHUNK;


typedef struct {
	int riff_chunk_size;
	FMTCHUNK  fmt;
	int data_chunk_size;
} RIFFWAVE;

#ifndef GUID_DEFINED
#define GUID_DEFINED
typedef struct _GUID
{
        unsigned long   Data1;
        unsigned short  Data2;
        unsigned short  Data3;
        unsigned char   Data4[8];
} GUID;
#endif  /* GUID_DEFINED */

#ifdef __cplusplus
extern "C" {
#endif

/* prototype */
int parse_riff_wave_header(FILE *, RIFFWAVE *);
int write_riff_wave_header(FILE *, RIFFWAVE *);
int update_riff_wave_header(FILE *, RIFFWAVE *);

#ifdef __cplusplus
}
#endif

#endif	/* _WAVEFILE_H */

/*******************************************************************************
        End of file
*******************************************************************************/
