/*

This software module was originally developed by

    Shusuke Takahashi (Sony Corporation)

    in the course of development of the MPEG-4 Audio standard (ISO/IEC 14496-3).
    This software module is an implementation of a part of one or more
    MPEG-4 Audio (ISO/IEC 14496-3) tools as specified by the MPEG-4 Audio
    standard (ISO/IEC 14496-3).
    ISO/IEC gives users of the MPEG-4 Audio standards (ISO/IEC 14496-3)
    free license to this software module or modifications thereof for use
    in hardware or software products claiming conformance to the MPEG-4
    Audio standards (ISO/IEC 14496-3).
    Those intending to use this software module in hardware or software
    products are advised that this use may infringe existing patents.
    The original developer of this software module and his/her company,
    the subsequent editors and their companies, and ISO/IEC have no
    liability for use of this software module or modifications thereof in
    an implementation.
    Copyright is not released for non MPEG-4 Audio (ISO/IEC 14496-3)
    conforming products. The original developer retains full right to use
    the code for his/her own purpose, assign or donate the code to a third
    party and to inhibit third party from using the code for non MPEG-4
    Audio (ISO/IEC 14496-3) conforming products.
    This copyright notice must be included in all copies or derivative works.

    Copyright (c)2015.

*/
#ifndef	_AS_COMPARE_H
#define _AS_COMPARE_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "libmpegassim.h"

/* macro definition */

#define	MPEG_AS_MIN_LENGTH_OF_VALID_ROUTE 5
#define	MPEG_AS_SCORESUM_THRESH	48
#define	MPEG_AS_NUM_BYTE_FEAT	16

#define	MPEG_AS_SEARCH_AVOIDANCE_WIDTH	15

#define	MPEG_AS_FEATURE_RESOLUTION_NOT_SET	0
#define	MPEG_AS_FEATURE_RESOLUTION_8_MSEC	8
#define	MPEG_AS_FEATURE_RESOLUTION_32_MSEC	32

#define	MPEG_AS_CALCULATION_MODE_0	0
#define	MPEG_AS_CALCULATION_MODE_1	1

#define	MPEG_AS_SAMPLING_RATE	8000

#define	MPEG_AS_COMPARE_RESOLUTION	32

#define	MPEG_AS_BLOCK_SIZE			64


#define	MPEG_AS_MAX_NUM_CANDIDATE	20

/* data type definition */
typedef	struct	_mpeg_as_sync_handle {
	int a_delay[MPEG_AS_MAX_NUM_CANDIDATE];
	float a_score[MPEG_AS_MAX_NUM_CANDIDATE];
	int num_delay;
	int	search_width;
	int	resolution0;
	int	resolution1;
	int	resolution_sim;
	
	unsigned char *p_buf0;
	unsigned char *p_buf1;
	unsigned char *p_buf_rate_convert0;
	unsigned char *p_buf_rate_convert1;
	int	len0;
	int len1;
	int	len_rate_convert0;
	int len_rate_convert1;
	float *p_scoresum;
	int *p_num_point;
	int	use_overlap_weighting;
	float framestep_new;
} HANDLE_AS_COMPARE_STRUCT;

/* proto type declaration */

void *mpeg_as_malloc(int size);
void *mpeg_as_calloc(int count, int size);
void mpeg_as_free(void *ptr);


int audio_sync_compare_features
(
MPEG_AS_SYNC_HANDLE handle,
const unsigned char *p_buf0,
const int len0,
const unsigned char *p_buf1,
const int len1
);

int audio_sync_estimate_time_difference
(
MPEG_AS_SYNC_HANDLE handle, 
float *p_maxval,
int	  *p_maxval_id
);


void audio_sync_pick_up_results
(
MPEG_AS_SYNC_HANDLE handle, 
float *p_score, 
int *p_delay_idx, 
float maxval, 
int maxval_id, 
int num_cand, 
int len0, 
int len1,
int *a_delay
);


extern const unsigned char g_a_popcnt[256];
#endif


/*****************************************************************************
	End of file
*****************************************************************************/
