/*

This software module was originally developed by

    Shusuke Takahashi (Sony Corporation)

    in the course of development of the MPEG-4 Audio standard (ISO/IEC 14496-3).
    This software module is an implementation of a part of one or more
    MPEG-4 Audio (ISO/IEC 14496-3) tools as specified by the MPEG-4 Audio
    standard (ISO/IEC 14496-3).
    ISO/IEC gives users of the MPEG-4 Audio standards (ISO/IEC 14496-3)
    free license to this software module or modifications thereof for use
    in hardware or software products claiming conformance to the MPEG-4
    Audio standards (ISO/IEC 14496-3).
    Those intending to use this software module in hardware or software
    products are advised that this use may infringe existing patents.
    The original developer of this software module and his/her company,
    the subsequent editors and their companies, and ISO/IEC have no
    liability for use of this software module or modifications thereof in
    an implementation.
    Copyright is not released for non MPEG-4 Audio (ISO/IEC 14496-3)
    conforming products. The original developer retains full right to use
    the code for his/her own purpose, assign or donate the code to a third
    party and to inhibit third party from using the code for non MPEG-4
    Audio (ISO/IEC 14496-3) conforming products.
    This copyright notice must be included in all copies or derivative works.

    Copyright (c)2015.

*/

#include	<stdlib.h>
#include	<string.h>

#include	"libmpegassim.h"
#include	"as_compare.h"

/* prototype */
static int audio_sync_rate_convert_feature(
const unsigned char *p_buf0_org, 
const int len0, 
const float feat_rate_convert_ratio, 
unsigned char *p_buf0);

/************************************************************************
	API functions
************************************************************************/
/* get handle */
MPEG_AS_SYNC_HANDLE	mpeg_as_sync_get_handle(void) 
{

	MPEG_AS_SYNC_HANDLE handle;


	handle = (MPEG_AS_SYNC_HANDLE)mpeg_as_calloc(1, (sizeof(HANDLE_AS_COMPARE_STRUCT)));
	if (handle == NULL) {
		return handle;
	}

	handle->search_width = 0;
	handle->resolution0 = MPEG_AS_FEATURE_RESOLUTION_NOT_SET;
	handle->resolution1 = MPEG_AS_FEATURE_RESOLUTION_NOT_SET;
	handle->resolution_sim = MPEG_AS_COMPARE_RESOLUTION;
	handle->use_overlap_weighting = 1;

	return handle;
}

/* init */ 
MPEG_AS_RESULT  mpeg_as_sync_init_handle (
MPEG_AS_SYNC_HANDLE  hData, 
char* p_buf_feature0, 
int size1, 
char* p_buf_feature1, 
int size2
 ) 
{
	MPEG_AS_RESULT result = MPEG_AS_S_OK;
	unsigned char *p_buf0 = (unsigned char *)p_buf_feature0;
	int datasize0 = size1;
	int	resolution0 = hData->resolution0;
	unsigned char *p_buf1 = (unsigned char *)p_buf_feature1;
	int datasize1 = size2;
	int	resolution1 = hData->resolution1;

	int	resolution_sim = hData->resolution_sim;

	int ret;
	float feat_rate_convert_ratio0;
	float feat_rate_convert_ratio1;

	if (hData == NULL) {
		fprintf(stderr, "mpeg_as_sync_init_handle argument error\n");
		return MPEG_AS_E_INVALID_HANDLE;
	} 

	if (p_buf_feature0 == NULL) {
		fprintf(stderr, "mpeg_as_sync_init_handle argument error\n");
		result = MPEG_AS_E_BAD_ARGUMENTS;
		goto FINISH_ROUTINE;
	}

	if (p_buf_feature1 == NULL) {
		fprintf(stderr, "mpeg_as_sync_init_handle argument error\n");
		result = MPEG_AS_E_BAD_ARGUMENTS;
		goto FINISH_ROUTINE;
	}

	if (size1 <= 0) {
		fprintf(stderr, "mpeg_as_sync_init_handle argument error: size1 <= 0\n");
		result = MPEG_AS_E_BAD_ARGUMENTS;
		goto FINISH_ROUTINE;
	}

	if (size2 <= 0) {
		fprintf(stderr, "mpeg_as_sync_init_handle argument error: size2 <= 0\n");
		result = MPEG_AS_E_BAD_ARGUMENTS;
		goto FINISH_ROUTINE;
	}

	if (resolution0 != MPEG_AS_FEATURE_RESOLUTION_8_MSEC && resolution0 != MPEG_AS_FEATURE_RESOLUTION_32_MSEC) {
		fprintf(stderr, "mpeg_as_sync_init_handle argument error\n");
		return MPEG_AS_E_BAD_ARGUMENTS;
	}

	if (resolution1 != MPEG_AS_FEATURE_RESOLUTION_8_MSEC && resolution1 != MPEG_AS_FEATURE_RESOLUTION_32_MSEC) {
		fprintf(stderr, "mpeg_as_sync_init_handle argument error\n");
		return MPEG_AS_E_BAD_ARGUMENTS;
	}

	if (datasize0 <= 0 || datasize1 <= 0) {
		fprintf(stderr, "illegal input file size\n");
		return MPEG_AS_E_SHORT_INPUT;
	}


	feat_rate_convert_ratio0 = (float)resolution_sim/resolution0;
	feat_rate_convert_ratio1 = (float)resolution_sim/resolution1;

	hData->len0 = datasize0/MPEG_AS_NUM_BYTE_FEAT;
	hData->len1 = datasize1/MPEG_AS_NUM_BYTE_FEAT;

	hData->p_buf0 = (unsigned char *)mpeg_as_malloc(datasize0);
	if (hData->p_buf0 == NULL) {
		fprintf(stderr, "malloc error\n");
		result = MPEG_AS_E_OUT_OF_MEMORY;
		goto FINISH_ROUTINE;
	}

	hData->p_buf1 = (unsigned char *)mpeg_as_malloc(datasize1);
	if (hData->p_buf1 == NULL) {
		fprintf(stderr, "malloc error\n");
		result = MPEG_AS_E_OUT_OF_MEMORY;
		goto FINISH_ROUTINE;
	}

	hData->p_buf_rate_convert0 = (unsigned char *)mpeg_as_calloc((int)(datasize0/feat_rate_convert_ratio0), sizeof(char));
	if (hData->p_buf_rate_convert0 == NULL) {
		fprintf(stderr, "malloc error\n");
		result = MPEG_AS_E_OUT_OF_MEMORY;
		goto FINISH_ROUTINE;
	}

	hData->p_buf_rate_convert1 = (unsigned char *)mpeg_as_calloc((int)(datasize1/feat_rate_convert_ratio1), sizeof(char));
	if (hData->p_buf_rate_convert1 == NULL) {
		fprintf(stderr, "malloc error\n");
		result = MPEG_AS_E_OUT_OF_MEMORY;
		goto FINISH_ROUTINE;
	}

#if _MSC_VER >=1500 
	memcpy_s(hData->p_buf0, datasize0, p_buf0, datasize0);
	memcpy_s(hData->p_buf1, datasize1, p_buf1, datasize1);
#else
	memcpy(hData->p_buf0, p_buf0, datasize0);
	memcpy(hData->p_buf1, p_buf1, datasize1);
#endif

	ret = audio_sync_rate_convert_feature(hData->p_buf0, hData->len0, feat_rate_convert_ratio0, hData->p_buf_rate_convert0);
	if (ret != MPEG_AS_S_OK) {
		fprintf(stderr, "error: audio_sync_rate_convert_feature\n");
		result = MPEG_AS_E_FAIL;
		goto FINISH_ROUTINE;
	}
	hData->len_rate_convert0 = (int)(hData->len0/feat_rate_convert_ratio0);

	ret = audio_sync_rate_convert_feature(hData->p_buf1, hData->len1, feat_rate_convert_ratio1, hData->p_buf_rate_convert1);
	if (ret != MPEG_AS_S_OK) {
		fprintf(stderr, "error: audio_sync_rate_convert_feature\n");
		result = MPEG_AS_E_FAIL;
		goto FINISH_ROUTINE;
	}
	hData->len_rate_convert1 = (int)(hData->len1/feat_rate_convert_ratio1);

	if (hData->len_rate_convert0 + hData->len_rate_convert1 - 1 <= 0) {
		fprintf(stderr, "input file size is too small\n");
		result = MPEG_AS_E_SHORT_INPUT;
		goto FINISH_ROUTINE;
	}

	hData->p_scoresum = (float *)mpeg_as_calloc(hData->len_rate_convert0 + hData->len_rate_convert1 - 1, sizeof(float));
	if (hData->p_scoresum == NULL) {
		fprintf(stderr, "malloc error\n");
		result = MPEG_AS_E_OUT_OF_MEMORY;
		goto FINISH_ROUTINE;
	}

	hData->p_num_point = (int *)mpeg_as_calloc(hData->len_rate_convert0 + hData->len_rate_convert1 - 1, sizeof(int));
	if (hData->p_num_point == NULL) {
		fprintf(stderr, "malloc error\n");
		result = MPEG_AS_E_OUT_OF_MEMORY;
		goto FINISH_ROUTINE;
	}

	hData->framestep_new = (float)(MPEG_AS_SAMPLING_RATE*resolution_sim/1000);



FINISH_ROUTINE:
	return result;

}


/* free */
MPEG_AS_RESULT  mpeg_as_sync_free_handle (
MPEG_AS_SYNC_HANDLE  hData
) 
{

	if (hData != NULL) {

		if (hData->p_buf0 != NULL) {
			mpeg_as_free(hData->p_buf0);			
		}

		if (hData->p_buf1 != NULL) {
			mpeg_as_free(hData->p_buf1);			
		}

		if (hData->p_buf_rate_convert0 != NULL) {
			mpeg_as_free(hData->p_buf_rate_convert0);			
		}

		if (hData->p_buf_rate_convert1 != NULL) {
			mpeg_as_free(hData->p_buf_rate_convert1);			
		}

		if (hData->p_scoresum != NULL) {
			mpeg_as_free(hData->p_scoresum);
		}

		if (hData->p_num_point != NULL) {
			mpeg_as_free(hData->p_num_point);			
		}

		mpeg_as_free(hData);
	}

	return MPEG_AS_S_OK;
}


/* get frame feature size */
MPEG_AS_RESULT  mpeg_as_sync_get_feature_size (
MPEG_AS_SYNC_HANDLE  hData, 
int*	p_frame_size
) 
{
	if (hData == NULL) {
		return MPEG_AS_E_INVALID_HANDLE;
	} 

	if (p_frame_size == NULL) {
		return MPEG_AS_E_BAD_ARGUMENTS;
	} 

	*p_frame_size = MPEG_AS_NUM_BYTE_FEAT;
	return MPEG_AS_S_OK;
}

/* get sampling rate */
MPEG_AS_RESULT  mpeg_as_sync_get_sampling_rate (
MPEG_AS_SYNC_HANDLE  hData, 
int* p_sampling_rate
) 
{
	if (hData == NULL) {
		return MPEG_AS_E_INVALID_HANDLE;
	} 

	if (p_sampling_rate == NULL) {
		return MPEG_AS_E_BAD_ARGUMENTS;
	} 

	*p_sampling_rate = MPEG_AS_SAMPLING_RATE;
	return MPEG_AS_S_OK;
}

/* set resolution */

MPEG_AS_RESULT  mpeg_as_sync_set_time_resolution (
 MPEG_AS_SYNC_HANDLE  hData, 
 int resolution0, 
 int resolution1,
 int resolution_sim
) 
{

	if (hData == NULL) {
		return MPEG_AS_E_INVALID_HANDLE;
	} 

	if (resolution0 != MPEG_AS_FEATURE_RESOLUTION_8_MSEC && resolution0 != MPEG_AS_FEATURE_RESOLUTION_32_MSEC) {
		return MPEG_AS_E_BAD_ARGUMENTS;
	}

	if (resolution1 != MPEG_AS_FEATURE_RESOLUTION_8_MSEC && resolution1 != MPEG_AS_FEATURE_RESOLUTION_32_MSEC) {
		return MPEG_AS_E_BAD_ARGUMENTS;
	}

	if (resolution_sim != MPEG_AS_FEATURE_RESOLUTION_8_MSEC && resolution_sim != MPEG_AS_FEATURE_RESOLUTION_32_MSEC) {
		return MPEG_AS_E_BAD_ARGUMENTS;
	}

	hData->resolution0 = resolution0;
	hData->resolution1 = resolution1;
	hData->resolution_sim = resolution_sim;

	return MPEG_AS_S_OK;
}

/* set calculation mode */
MPEG_AS_RESULT  mpeg_as_sync_set_calculation_mode (
 MPEG_AS_SYNC_HANDLE  hData, 
 int calculation_mode
) 
{

	if (hData == NULL) {
		return MPEG_AS_E_INVALID_HANDLE;
	} 

	if (calculation_mode != MPEG_AS_CALCULATION_MODE_0 && calculation_mode != MPEG_AS_CALCULATION_MODE_1) {
		return MPEG_AS_E_BAD_ARGUMENTS;
	}

	hData->use_overlap_weighting = calculation_mode;

	return MPEG_AS_S_OK;
}

/* set search area */
MPEG_AS_RESULT  mpeg_as_sync_set_search_area (
 MPEG_AS_SYNC_HANDLE  hData, 
 int search_area
) 
{
	int result = MPEG_AS_S_OK;
	
	if (hData == NULL) {
		return MPEG_AS_E_INVALID_HANDLE;
	} 

	if (search_area <= 0) {
		return MPEG_AS_E_BAD_ARGUMENTS;
	}
	
	hData->search_width = search_area;

	return result;
}

/* synchronization */
MPEG_AS_RESULT  mpeg_as_sync_processing (
MPEG_AS_SYNC_HANDLE  hData
) 
{
	int result = MPEG_AS_S_OK;
	MPEG_AS_SYNC_HANDLE handle = hData;
	int *p_delay = hData->a_delay;
	float *p_score = hData->a_score;
	int *p_num_delay = &(hData->num_delay);

	int ret;
	float maxval;
	int maxval_id;
	int num_score_fine = 0;
	int a_delay[MPEG_AS_MAX_NUM_CANDIDATE];
	float a_score[MPEG_AS_MAX_NUM_CANDIDATE];
	int a_delay_idx[MPEG_AS_MAX_NUM_CANDIDATE];
	int idx_cand;
	int num_cand = MPEG_AS_MAX_NUM_CANDIDATE;
	int len0;
	int len1;
	if (hData == NULL) {
		result = MPEG_AS_E_INVALID_HANDLE;
		goto FINISH_ROUTINE;
	}

	len0 = handle->len_rate_convert0;
	len1 = handle->len_rate_convert1;


	if (handle->len0 - MPEG_AS_BLOCK_SIZE + 1 <= 0) {
		fprintf(stderr, "input file size is too small\n");
		return MPEG_AS_E_SHORT_INPUT;
	}


	if (handle->len1 - MPEG_AS_BLOCK_SIZE + 1 <= 0) {
		fprintf(stderr, "input file size is too small\n");
		return MPEG_AS_E_SHORT_INPUT;
	}
	/* -------------------------------------------------
	*  calculate similarity
	* -------------------------------------------------*/
	ret = audio_sync_compare_features(handle, handle->p_buf_rate_convert0, handle->len_rate_convert0, handle->p_buf_rate_convert1, handle->len_rate_convert1);
	if (ret != MPEG_AS_S_OK) {
		fprintf(stderr, "audio_sync_compare_features error\n");
		result = ret;
		goto FINISH_ROUTINE;
	}

	/* -------------------------------------------------
	*  estimate time difference
	* -------------------------------------------------*/
	ret = audio_sync_estimate_time_difference(handle, &maxval, &maxval_id);
	if (ret != MPEG_AS_S_OK) {
		fprintf(stderr, "audio_sync_estimate_time_difference error\n");
		result = ret;
		goto FINISH_ROUTINE;
	}

	/* -------------------------------------------------
	*  pick up candidates
	* -------------------------------------------------*/
	num_cand = MPEG_AS_MAX_NUM_CANDIDATE;
	audio_sync_pick_up_results(handle, a_score, a_delay_idx, maxval, maxval_id, num_cand, len0, len1, a_delay);
	num_score_fine = num_cand;


	/* -------------------------------------------------
	*  calculate scores
	* -------------------------------------------------*/
	for (idx_cand = 0; idx_cand < num_score_fine; idx_cand++) {

		p_delay[idx_cand] = a_delay[idx_cand];
		p_score[idx_cand] = a_score[idx_cand];
	}


	*p_num_delay = num_score_fine;


FINISH_ROUTINE:
	return result;
}

/* get maximum number of result */
MPEG_AS_RESULT  mpeg_as_sync_get_max_num_result (
MPEG_AS_SYNC_HANDLE hData, 
int* p_max_num_result
)
{

	if (hData == NULL) {
		return MPEG_AS_E_INVALID_HANDLE;
	} 

	if (p_max_num_result == NULL) {
		return MPEG_AS_E_BAD_ARGUMENTS;
	} 

	*p_max_num_result = MPEG_AS_MAX_NUM_CANDIDATE;

	return MPEG_AS_S_OK;

}

/* get result */
MPEG_AS_RESULT  mpeg_as_sync_get_result (
MPEG_AS_SYNC_HANDLE  hData, 
int num_result, 
int *p_delay, 
double *p_score
)
{
	MPEG_AS_RESULT result = MPEG_AS_S_OK;

	int idx_result;

	if (hData == NULL) {
		return MPEG_AS_E_INVALID_HANDLE;
	} 

	if (num_result > MPEG_AS_MAX_NUM_CANDIDATE || num_result < 0) {
		return MPEG_AS_E_BAD_ARGUMENTS;
	}

	if (p_delay == NULL) {
		return MPEG_AS_E_BAD_ARGUMENTS;
	} 

	for (idx_result = 0; idx_result < num_result; idx_result++) {
		if (idx_result < hData->num_delay) {
			p_delay[idx_result] = hData->a_delay[idx_result];
			p_score[idx_result] = hData->a_score[idx_result];
		} else {
			p_delay[idx_result] = 0;
			p_score[idx_result] = -1.0;
		}
	}

	return result;
}



void *mpeg_as_malloc(int size) {
	void *ptr;

	ptr = malloc(size);

	return ptr;
}

void *mpeg_as_calloc(int count, int size) {
	void *ptr;
	ptr = calloc(count, size);

	return ptr;
}

void mpeg_as_free(void *ptr) {
	
	free(ptr);

	return;
}

/* sub functions */

/* down convert features */
static int audio_sync_rate_convert_feature(
const unsigned char *p_buf0_org, 
const int len0, 
const float feat_rate_convert_ratio, 
unsigned char *p_buf0
)
{
	int idx_frame0;
	int idx_byte;
	int idx_frame0_rate_convert = 0;
	if (feat_rate_convert_ratio >= 1.0) {
		/* downward */
		for (idx_frame0 = 0; idx_frame0 < len0 - ((int)feat_rate_convert_ratio - 1); idx_frame0 += (int)feat_rate_convert_ratio) {
			int idx_frame0_in = 0;
			char feat_buf[MPEG_AS_NUM_BYTE_FEAT];
			memset(feat_buf, 0, sizeof(char)*MPEG_AS_NUM_BYTE_FEAT);

			for (idx_frame0_in = 0; idx_frame0_in < feat_rate_convert_ratio; idx_frame0_in++) {
				for (idx_byte = 0; idx_byte < MPEG_AS_NUM_BYTE_FEAT; idx_byte++) {
					feat_buf[idx_byte] |= p_buf0_org[(idx_frame0 + idx_frame0_in)*MPEG_AS_NUM_BYTE_FEAT + idx_byte];
				}
			}

			for (idx_byte = 0; idx_byte < MPEG_AS_NUM_BYTE_FEAT; idx_byte++) {
				p_buf0[idx_frame0_rate_convert*MPEG_AS_NUM_BYTE_FEAT + idx_byte] =
					feat_buf[idx_byte];
			}

			idx_frame0_rate_convert++;
		} 
	} else {
		/* upward */
		for (idx_frame0 = 0; idx_frame0 < len0; idx_frame0 ++) {
			int idx_frame0_in = 0;

			for (idx_frame0_in = 0; idx_frame0_in < 1/feat_rate_convert_ratio; idx_frame0_in++) {
				for (idx_byte = 0; idx_byte < MPEG_AS_NUM_BYTE_FEAT; idx_byte++) {
					p_buf0[idx_frame0_rate_convert*MPEG_AS_NUM_BYTE_FEAT + idx_byte] =
											p_buf0_org[idx_frame0*MPEG_AS_NUM_BYTE_FEAT + idx_byte];
				}
				idx_frame0_rate_convert++;
			}
		} 
	}

	return MPEG_AS_S_OK;
}


/*****************************************************************************
	End of file
*****************************************************************************/
