/*

This software module was originally developed by

    Shusuke Takahashi (Sony Corporation)

    in the course of development of the MPEG-4 Audio standard (ISO/IEC 14496-3).
    This software module is an implementation of a part of one or more
    MPEG-4 Audio (ISO/IEC 14496-3) tools as specified by the MPEG-4 Audio
    standard (ISO/IEC 14496-3).
    ISO/IEC gives users of the MPEG-4 Audio standards (ISO/IEC 14496-3)
    free license to this software module or modifications thereof for use
    in hardware or software products claiming conformance to the MPEG-4
    Audio standards (ISO/IEC 14496-3).
    Those intending to use this software module in hardware or software
    products are advised that this use may infringe existing patents.
    The original developer of this software module and his/her company,
    the subsequent editors and their companies, and ISO/IEC have no
    liability for use of this software module or modifications thereof in
    an implementation.
    Copyright is not released for non MPEG-4 Audio (ISO/IEC 14496-3)
    conforming products. The original developer retains full right to use
    the code for his/her own purpose, assign or donate the code to a third
    party and to inhibit third party from using the code for non MPEG-4
    Audio (ISO/IEC 14496-3) conforming products.
    This copyright notice must be included in all copies or derivative works.

    Copyright (c)2015.

*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "libmpegassim.h"

#define MPEG_AS_OK 0
#define MPEG_AS_NG 1

static float s_startpos_1st = 0;
static float s_duration_1st = 0;
static float s_startpos_2nd = 0;
static float s_duration_2nd = 0;
static float s_search_area = 0;
static int s_resolution_sim = 0;
static int s_calculation_mode = 1;
static int s_num_result = 1;
static char  *s_filename_1st = NULL;
static char  *s_filename_2nd = NULL;
static char  *s_filename_result = NULL;

static int get_option(int, char **);
static void show_usage();
static int read_audio_specific_config(FILE *fi_1st, int *p_feature_type, int *p_frame_length, int *p_resolution);

int main(int argc, char* argv[])
{

	MPEG_AS_SYNC_HANDLE	handle_sync = NULL;	

	int feature_size;
	int sampling_freq;

	char *p_buf_feature_1st = NULL;
	char *p_buf_feature_2nd = NULL;

	int num_result;
	int max_num_result;

	int *p_delay = NULL;
	double *p_score = NULL;

	MPEG_AS_RESULT	result;
	int ret = MPEG_AS_OK;

	FILE *fi_1st = NULL;
	FILE *fi_2nd = NULL;

	FILE *fout = NULL;

	int filesize_1st;
	int filesize_2nd;

	int startpos_byte_1st;
	int endpos_byte_1st;
	int startpos_byte_2nd;
	int endpos_byte_2nd;

	int	nbyte;

	int	feature_type_1st;
	int	feature_type_2nd;
	int	frame_length_1st;
	int	frame_length_2nd;
	int	resolution_1st;
	int	resolution_2nd;
	int	header_size_1st;
	int	header_size_2nd;


	double framestep_sec_1st;
	double framestep_sec_2nd;

	int idx_result;

	fprintf(stderr, " -- MPEG Audio Synchronization (version 1.0) --\n");

	if (get_option(argc, argv) != 0) {
		show_usage();
		ret = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}

	/* get handle */
	handle_sync = mpeg_as_sync_get_handle();
	if (handle_sync == NULL) {
		fprintf(stderr, "mpeg_as_sync_get_handle() error\n");
		ret = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}

	/* get parameters */
	result = mpeg_as_sync_get_feature_size(handle_sync, &feature_size);
	if (MPEG_AS_FAILED(result)) {
		printf("mpeg_as_sync_get_feature_size error\n");
		ret = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}

	result = mpeg_as_sync_get_sampling_rate(handle_sync, &sampling_freq);
	if (MPEG_AS_FAILED(result)) {
		printf("mpeg_as_sync_get_sampling_rate error\n");
		ret = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}


	if (s_filename_1st == NULL || s_filename_2nd == NULL || s_filename_result == NULL) {
		fprintf(stderr, "invalid filename\n");
		ret = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}

#if _MSC_VER >=1500 
	result = fopen_s(&fi_1st, s_filename_1st, "rb");
	if (result != 0) {
		fprintf(stderr, "can't open file\n");
		ret = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}
#else
	fi_1st = fopen(s_filename_1st, "rb");
	if (fi_1st == NULL) {
		fprintf(stderr, "can't open file\n");
		ret = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}
#endif


#if _MSC_VER >=1500 
	result = fopen_s(&fi_2nd, s_filename_2nd, "rb");
	if (result != 0) {
		fprintf(stderr, "can't open file\n");
		ret = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}
#else
	fi_2nd = fopen(s_filename_2nd, "rb");
	if (fi_2nd == NULL) {
		fprintf(stderr, "can't open file\n");
		ret = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}
#endif

#if _MSC_VER >=1500 
	result = fopen_s(&fout, s_filename_result, "w");
	if (result != 0) {
		fprintf(stderr, "can't open file\n");
		ret = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}
#else
	fout = fopen(s_filename_result, "w");
	if (fout == NULL) {
		fprintf(stderr, "can't open file\n");
		ret = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}
#endif

	/* read AudioSyncFeatureSpecificConfig() */
	result = read_audio_specific_config(fi_1st, &feature_type_1st, &frame_length_1st ,&resolution_1st);
	if (result != MPEG_AS_OK) {
		fprintf(stderr, "read_audio_specific_config error\n");
		ret = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}
	header_size_1st = ftell(fi_1st);

	result = read_audio_specific_config(fi_2nd, &feature_type_2nd, &frame_length_2nd, &resolution_2nd);
	if (result != MPEG_AS_OK) {
		fprintf(stderr, "read_audio_specific_config error\n");
		ret = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}
	header_size_2nd = ftell(fi_2nd);

	/* get file size */
	fseek(fi_1st, 0, SEEK_END);
	filesize_1st = ftell(fi_1st) - header_size_1st;

	fseek(fi_2nd, 0, SEEK_END);
	filesize_2nd = ftell(fi_2nd) - header_size_2nd;

	startpos_byte_1st = header_size_1st;
	endpos_byte_1st = header_size_1st + filesize_1st;
	startpos_byte_2nd = header_size_2nd;
	endpos_byte_2nd = header_size_2nd + filesize_2nd;


	framestep_sec_1st = (double)resolution_1st*0.001;
	framestep_sec_2nd = (double)resolution_2nd*0.001;

	if (s_startpos_1st != 0.0) {
		startpos_byte_1st = feature_size*(int)(s_startpos_1st/framestep_sec_1st) + header_size_1st;
	}

	if (s_duration_1st != 0.0) {
		endpos_byte_1st = startpos_byte_1st + feature_size*(int)(s_duration_1st/framestep_sec_1st);
	}

	if (s_startpos_2nd != 0.0) {
		startpos_byte_2nd = feature_size*(int)(s_startpos_2nd/framestep_sec_2nd) + header_size_2nd;
	}
	
	if (s_duration_2nd != 0.0) {
		endpos_byte_2nd = startpos_byte_2nd + feature_size*(int)(s_duration_2nd/framestep_sec_2nd);
	}

	fseek(fi_1st, startpos_byte_1st, SEEK_SET);
	fseek(fi_2nd, startpos_byte_2nd, SEEK_SET);

	/* update filesize  */
	filesize_1st = endpos_byte_1st - startpos_byte_1st;
	filesize_2nd = endpos_byte_2nd - startpos_byte_2nd;

	p_buf_feature_1st = (char *)malloc(filesize_1st);
	if (p_buf_feature_1st == NULL) {
		fprintf(stderr, "malloc error\n");
		result = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}

	p_buf_feature_2nd = (char *)malloc(filesize_2nd);
	if (p_buf_feature_2nd == NULL) {
		fprintf(stderr, "malloc error\n");
		result = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}

	nbyte = fread(p_buf_feature_1st, sizeof(char), filesize_1st, fi_1st);
	if (nbyte != filesize_1st) {
		fprintf(stderr, "warning: invalid duration for inputfile1\n");
		filesize_1st = nbyte;
	}

	nbyte = fread(p_buf_feature_2nd, sizeof(char), filesize_2nd, fi_2nd);
	if (nbyte != filesize_2nd) {
		fprintf(stderr, "warning: invalid duration for inputfile2\n");
		filesize_2nd = nbyte;
	}

	if (s_resolution_sim == 0) {
		/* default setting  */
		s_resolution_sim = 32;
	}	

	result = mpeg_as_sync_set_time_resolution(handle_sync, resolution_1st, resolution_2nd, s_resolution_sim);
	if (MPEG_AS_FAILED(result)) {
		fprintf(stderr, "mpeg_as_sync_set_time_resolution error\n");
		ret = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}

	result = mpeg_as_sync_set_calculation_mode(handle_sync, s_calculation_mode);
	if (MPEG_AS_FAILED(result)) {
		fprintf(stderr, "mpeg_as_sync_set_calculation_mode error\n");
		ret = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}

	if (s_search_area != 0.0) {
		result = mpeg_as_sync_set_search_area(handle_sync, (int)(s_search_area*sampling_freq));
		if (MPEG_AS_FAILED(result)) {
			fprintf(stderr, "mpeg_as_sync_set_search_area error\n");
			ret = MPEG_AS_NG;
			goto FINISH_ROUTINE;
		}
	}

	result = mpeg_as_sync_init_handle(handle_sync, p_buf_feature_1st, filesize_1st, p_buf_feature_2nd, filesize_2nd);
	if (MPEG_AS_FAILED(result)) {
		fprintf(stderr, "mpeg_as_sync_init_handle error\n");
		ret = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}

	fprintf(stderr, " 1st file: \n");
	fprintf(stderr, "    Filename:       %s\n", s_filename_1st);
	fprintf(stderr, "    Resolution:     %d msec\n", resolution_1st);
	fprintf(stderr, "    Start position: %f sec\n", (startpos_byte_1st - header_size_1st)/feature_size*resolution_1st*0.001);
	fprintf(stderr, "    End position:   %f sec\n", (endpos_byte_1st - header_size_1st)/feature_size*resolution_1st*0.001);
	fprintf(stderr, "\n");
	fprintf(stderr, " 2nd file: \n");
	fprintf(stderr, "    Filename:       %s\n", s_filename_2nd);
	fprintf(stderr, "    Resolution:     %d msec\n", resolution_2nd);
	fprintf(stderr, "    Start position: %f sec\n", (startpos_byte_2nd - header_size_2nd)/feature_size*resolution_2nd*0.001);
	fprintf(stderr, "    End position:   %f sec\n", (endpos_byte_2nd - header_size_2nd)/feature_size*resolution_2nd*0.001);
	fprintf(stderr, "\n");
	fprintf(stderr, " Similarity calculation: \n");
	fprintf(stderr, "    Filename:       %s\n", s_filename_result);
	fprintf(stderr, "    Resolution:     %d msec\n", s_resolution_sim);
	if (s_search_area != 0.0) {
		fprintf(stderr, "    Search area:    [-%f, %f]\n", s_search_area, s_search_area);
	} else {
		fprintf(stderr, "    Search area:    all\n");
	}

	result = mpeg_as_sync_processing(handle_sync);
	if (MPEG_AS_FAILED(result)) {
		fprintf(stderr, "mpeg_as_sync_processing error %x\n", result);
		ret = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}

	result = mpeg_as_sync_get_max_num_result(handle_sync, &max_num_result);
	if (MPEG_AS_FAILED(result)) {
		fprintf(stderr, "mpeg_as_sync_get_max_num_result error\n");
		ret = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}

	p_delay = (int *)malloc(sizeof(int)*max_num_result);
	if (p_delay == NULL) {
		fprintf(stderr, "malloc error (p_delay)\n");
		result = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}

	p_score = (double *)malloc(sizeof(double)*max_num_result);
	if (p_score == NULL) {
		fprintf(stderr, "malloc error (p_score)\n");
		result = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}

	num_result = s_num_result;
	if (num_result > max_num_result) {
		num_result = max_num_result;
	}

	result = mpeg_as_sync_get_result(handle_sync, num_result, p_delay, p_score);
	if (MPEG_AS_FAILED(result)) {
		fprintf(stderr, "mpeg_as_sync_get_result error\n");
		ret = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}

	fprintf(stderr, "----------------------------\n");
	fprintf(stderr, "delay(sec)\tscore\n");
	fprintf(stderr, "----------------------------\n");
	for (idx_result = 0; idx_result < num_result; idx_result++) {
		if ( p_score[idx_result] != 0.0) {
			fprintf(stderr, "%f\t%f\n", (float)p_delay[idx_result]/sampling_freq, p_score[idx_result]);
		}
	}

	fprintf(fout, "%f\n", (float)p_delay[0]/sampling_freq);

FINISH_ROUTINE:

	if (handle_sync != NULL) {
		result = mpeg_as_sync_free_handle(handle_sync);
	}

	if (p_buf_feature_1st != NULL) {
		free(p_buf_feature_1st);
	}

	if (p_buf_feature_2nd != NULL) {
		free(p_buf_feature_2nd);
	}

	if (p_delay != NULL) {
		free(p_delay);
	}

	if (p_score != NULL) {
		free(p_score);
	}

	if (fi_1st != NULL) {
		fclose(fi_1st);
	}

	if (fi_2nd != NULL) {
		fclose(fi_2nd);
	}

	if (fout != NULL) {
		fclose(fout);
	}

	return ret;

}

static int get_option(
	int			argc, 
	char		**argv
)
{
    int		i, j;
    char	*p_ptr;

	for (i = 1, j = 0; i < argc; i++) {
		p_ptr = argv[i];
		if (p_ptr[0] == '-' && p_ptr[1] != 0) {
			p_ptr++;
			if (0) {
				;
			} else if (!strcmp(p_ptr, "h")) {
				return 1;		
			} else if (!strcmp(p_ptr, "ss1")) {
				s_startpos_1st = (float)atof(argv[++i]);
			} else if (!strcmp(p_ptr, "ss2")) {
				s_startpos_2nd = (float)atof(argv[++i]);
			} else if (!strcmp(p_ptr, "ds1")) {
				s_duration_1st = (float)atof(argv[++i]);
			} else if (!strcmp(p_ptr, "ds2")) {
				s_duration_2nd = (float)atof(argv[++i]);
			} else if (!strcmp(p_ptr, "as")) {
				s_search_area = (float)atof(argv[++i]);
			} else if (!strcmp(p_ptr, "n")) {
				s_num_result = (int)atoi(argv[++i]);
			} else if (!strcmp(p_ptr, "r")) {
				s_resolution_sim = (int)atoi(argv[++i]);
			} else if (!strcmp(p_ptr, "m")) {
				s_calculation_mode = (int)atoi(argv[++i]);
			} else {
				fprintf(stderr, "-%s : unknown option\n", p_ptr);
				return 1;
			}
		} else {
			if (j == 0) {
				s_filename_1st = p_ptr;
			} else if (j == 1) {
				s_filename_2nd = p_ptr;
			} else if (j == 2) {
				s_filename_result = p_ptr;
			} 
			j++;
		}
    }

	if (j != 3) {
		return 1;
	}
	return 0;
}

static void show_usage() {

	fprintf(stderr, "\nusage: mpeg_as_sim [options] inputfile1 inputfile2 outputfile\n");
	fprintf(stderr, "options:\n");
	fprintf(stderr, "   -ss1: set start point in sec for inputfile1\n");
	fprintf(stderr, "   -ds1: set duration in sec for inputfile1\n");
	fprintf(stderr, "   -ss2: set start point in sec for inputfile2\n");
	fprintf(stderr, "   -ds2: set duration in sec for inputfile2\n");
	fprintf(stderr, "   -as : set search area in sec :  calculation is done only for time difference of [-as, as].\n");
	fprintf(stderr, "   -r  : set time resolution for similarity calculation (8msec or 32msec)\n");
	fprintf(stderr, "   -m  : set similarity calculation mode: 0 or 1(default)\n");
	fprintf(stderr, "   -n  : set number of candidates to be shown\n");

	return;

}

static int read_audio_specific_config(FILE *fi_1st, int *p_feature_type, int *p_frame_length, int *p_resolution)
{
	unsigned char a_audio_specific_config[4];
	unsigned int nbyte;
	unsigned int audio_sync_feature_type;
	unsigned int audio_sync_feature_frame_length_index;
	unsigned int audio_sync_feature_frame_length;
	unsigned int audio_sync_feature_time_resolution_index;
	unsigned int number_of_streams_index;
	unsigned int audio_specific_config_size = 4;
	int	result = MPEG_AS_OK;

	nbyte = fread(a_audio_specific_config, sizeof(char), audio_specific_config_size, fi_1st);
	if (nbyte != 4) {
		fprintf(stderr, "fread error\n");
		result = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}

	audio_sync_feature_type = a_audio_specific_config[0] >> 4;
	audio_sync_feature_frame_length_index = 0x0f & a_audio_specific_config[0];
	audio_sync_feature_time_resolution_index = a_audio_specific_config[1] >> 4;
	number_of_streams_index = 0x0f & a_audio_specific_config[1];

	if (audio_sync_feature_type != 0) {
		fprintf(stderr, "illegal audio_sync_feature_type (%d)\n", audio_sync_feature_type);
		result = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}
	*p_feature_type = 0;

	if (audio_sync_feature_frame_length_index == 0) {
		*p_frame_length = audio_sync_feature_frame_length = 128;
	} else {
		fprintf(stderr, "illegal audio_sync_feature_frame_length_index (%d)\n", audio_sync_feature_frame_length_index);
		result = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}

	if (audio_sync_feature_time_resolution_index == 0) {
		*p_resolution = 32;
	} else if (audio_sync_feature_time_resolution_index == 1) {
		*p_resolution = 8;
	} else {
		fprintf(stderr, "illegal audio_sync_feature_time_resolution_index (%d)\n", audio_sync_feature_time_resolution_index);
		result = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}

	if (number_of_streams_index != 1) {
		fprintf(stderr, "illegal number_of_streams_index (%d)\n", number_of_streams_index);
		result = MPEG_AS_NG;
		goto FINISH_ROUTINE;
	}
FINISH_ROUTINE:
	return result;
}
/*******************************************************************************
	End of file
*******************************************************************************/
